import unittest
import requests

from httpretty import HTTPretty

from ...utils import module_member, parse_qs, PARTIAL_TOKEN_SESSION_NAME
from ...backends.utils import user_backends_data, load_backends
from ..strategy import TestStrategy
from ..models import User, TestUserSocialAuth, TestNonce, \
    TestAssociation, TestCode, TestStorage


class BaseBackendTest(unittest.TestCase):
    backend = None
    backend_path = None
    name = None
    complete_url = ''
    raw_complete_url = '/complete/{0}'

    def setUp(self):
        HTTPretty.enable()
        Backend = module_member(self.backend_path)
        self.strategy = TestStrategy(TestStorage)
        self.backend = Backend(self.strategy, redirect_uri=self.complete_url)
        self.name = self.backend.name.upper().replace('-', '_')
        self.complete_url = self.strategy.build_absolute_uri(
            self.raw_complete_url.format(self.backend.name)
        )
        backends = (self.backend_path,
                    'social_core.tests.backends.test_broken.BrokenBackendAuth')
        self.strategy.set_settings({
            'SOCIAL_AUTH_AUTHENTICATION_BACKENDS': backends
        })
        self.strategy.set_settings(self.extra_settings())
        # Force backends loading to trash PSA cache
        load_backends(backends, force_load=True)
        User.reset_cache()
        TestUserSocialAuth.reset_cache()
        TestNonce.reset_cache()
        TestAssociation.reset_cache()
        TestCode.reset_cache()

    def tearDown(self):
        HTTPretty.disable()
        self.backend = None
        self.strategy = None
        self.name = None
        self.complete_url = None
        User.reset_cache()
        TestUserSocialAuth.reset_cache()
        TestNonce.reset_cache()
        TestAssociation.reset_cache()
        TestCode.reset_cache()

    def extra_settings(self):
        return {}

    def do_start(self):
        raise NotImplementedError('Implement in subclass')

    def do_login(self):
        user = self.do_start()
        username = self.expected_username
        self.assertEqual(user.username, username)
        self.assertEqual(self.strategy.session_get('username'), username)
        self.assertEqual(self.strategy.get_user(user.id), user)
        self.assertEqual(self.backend.get_user(user.id), user)
        user_backends = user_backends_data(
            user,
            self.strategy.get_setting('SOCIAL_AUTH_AUTHENTICATION_BACKENDS'),
            self.strategy.storage
        )
        self.assertEqual(len(list(user_backends.keys())), 3)
        self.assertEqual('associated' in user_backends, True)
        self.assertEqual('not_associated' in user_backends, True)
        self.assertEqual('backends' in user_backends, True)
        self.assertEqual(len(user_backends['associated']), 1)
        self.assertEqual(len(user_backends['not_associated']), 1)
        self.assertEqual(len(user_backends['backends']), 2)
        return user

    def pipeline_settings(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_PIPELINE': (
                'social_core.pipeline.social_auth.social_details',
                'social_core.pipeline.social_auth.social_uid',
                'social_core.pipeline.social_auth.auth_allowed',
                'social_core.tests.pipeline.ask_for_password',
                'social_core.tests.pipeline.ask_for_slug',
                'social_core.pipeline.social_auth.social_user',
                'social_core.pipeline.user.get_username',
                'social_core.pipeline.social_auth.associate_by_email',
                'social_core.pipeline.user.create_user',
                'social_core.pipeline.social_auth.associate_user',
                'social_core.pipeline.social_auth.load_extra_data',
                'social_core.tests.pipeline.set_password',
                'social_core.tests.pipeline.set_slug',
                'social_core.pipeline.user.user_details'
            )
        })

    def pipeline_handlers(self, url):
        HTTPretty.register_uri(HTTPretty.GET, url, status=200, body='foobar')
        HTTPretty.register_uri(HTTPretty.POST, url, status=200)

    def pipeline_password_handling(self, url):
        password = 'foobar'
        requests.get(url)
        requests.post(url, data={'password': password})

        data = parse_qs(HTTPretty.last_request.body)
        self.assertEqual(data['password'], password)
        self.strategy.session_set('password', data['password'])
        return password

    def pipeline_slug_handling(self, url):
        slug = 'foo-bar'
        requests.get(url)
        requests.post(url, data={'slug': slug})

        data = parse_qs(HTTPretty.last_request.body)
        self.assertEqual(data['slug'], slug)
        self.strategy.session_set('slug', data['slug'])
        return slug

    def do_partial_pipeline(self):
        url = self.strategy.build_absolute_uri('/password')
        self.pipeline_settings()
        redirect = self.do_start()
        self.assertEqual(redirect.url, url)
        self.pipeline_handlers(url)

        password = self.pipeline_password_handling(url)
        token = self.strategy.session_pop(PARTIAL_TOKEN_SESSION_NAME)
        partial = self.strategy.partial_load(token)
        self.assertEqual(partial.backend, self.backend.name)
        redirect = self.backend.continue_pipeline(partial)

        url = self.strategy.build_absolute_uri('/slug')
        self.assertEqual(redirect.url, url)
        self.pipeline_handlers(url)
        slug = self.pipeline_slug_handling(url)

        token = self.strategy.session_pop(PARTIAL_TOKEN_SESSION_NAME)
        partial = self.strategy.partial_load(token)
        self.assertEqual(partial.backend, self.backend.name)
        user = self.backend.continue_pipeline(partial)

        self.assertEqual(user.username, self.expected_username)
        self.assertEqual(user.slug, slug)
        self.assertEqual(user.password, password)
        return user



"""
Behance OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/behance.html
"""
from .oauth import BaseOAuth2


class BehanceOAuth2(BaseOAuth2):
    """Behance OAuth authentication backend"""
    name = 'behance'
    AUTHORIZATION_URL = 'https://www.behance.net/v2/oauth/authenticate'
    ACCESS_TOKEN_URL = 'https://www.behance.net/v2/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = '|'
    EXTRA_DATA = [('username', 'username')]
    REDIRECT_STATE = False

    def get_user_id(self, details, response):
        return response['user']['id']

    def get_user_details(self, response):
        """Return user details from Behance account"""
        user = response['user']
        fullname, first_name, last_name = self.get_user_names(
            user['display_name'], user['first_name'], user['last_name']
        )
        return {'username': user['username'],
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': ''}

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        # Pull up the embedded user attributes so they can be found as extra
        # data. See the example token response for possible attributes:
        # http://www.behance.net/dev/authentication#step-by-step
        data = response.copy()
        data.update(response['user'])
        return super().extra_data(user, uid, data, details, *args, **kwargs)



import json

from .oauth import OAuth2Test


class CognitoAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.cognito.CognitoOAuth2'
    pool_domain = 'https://social_core.auth.eu-west-1.amazoncognito.com'
    expected_username = 'cognito.account.ABCDE1234'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'given_name': 'John',
        'family_name': 'Doe',
        'username': 'cognito.account.ABCDE1234',
        'email': 'john@doe.test',
    })

    @property
    def user_data_url(self):
        return self.backend.user_data_url()

    def extra_settings(self):
        settings = super().extra_settings()
        settings.update({
            'SOCIAL_AUTH_' + self.name + '_POOL_DOMAIN': self.pool_domain,
        })
        return settings

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Twitter OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/twitter.html
"""
from .oauth import BaseOAuth1
from ..exceptions import AuthCanceled


class TwitterOAuth(BaseOAuth1):
    """Twitter OAuth authentication backend"""
    name = 'twitter'
    EXTRA_DATA = [('id', 'id')]
    REQUEST_TOKEN_METHOD = 'POST'
    ACCESS_TOKEN_METHOD = 'POST'
    AUTHORIZATION_URL = 'https://api.twitter.com/oauth/authenticate'
    REQUEST_TOKEN_URL = 'https://api.twitter.com/oauth/request_token'
    ACCESS_TOKEN_URL = 'https://api.twitter.com/oauth/access_token'
    REDIRECT_STATE = True

    def process_error(self, data):
        if 'denied' in data:
            raise AuthCanceled(self)
        else:
            super().process_error(data)

    def get_user_details(self, response):
        """Return user details from Twitter account"""
        fullname, first_name, last_name = self.get_user_names(response['name'])
        return {'username': response['screen_name'],
                'email': response.get('email', ''),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        return self.get_json(
            'https://api.twitter.com/1.1/account/verify_credentials.json',
            params={'include_email': 'true'},
            auth=self.oauth_auth(access_token)
        )



import datetime
from urllib.parse import urlencode

from httpretty import HTTPretty

from ...exceptions import AuthMissingParameter

from .open_id import OpenIdTest


JANRAIN_NONCE = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')


class LiveJournalOpenIdTest(OpenIdTest):
    backend_path = 'social_core.backends.livejournal.LiveJournalOpenId'
    expected_username = 'foobar'
    discovery_body = ''.join([
        '<xrds:XRDS xmlns:xrds="xri://$xrds" xmlns="xri://$xrd*($v*2.0)">',
        '<XRD>',
        '<Service priority="0">',
        '<Type>http://specs.openid.net/auth/2.0/signon</Type>',
        '<URI>http://www.livejournal.com/openid/server.bml</URI>',
        '<LocalID>http://foobar.livejournal.com/</LocalID>',
        '</Service>',
        '</XRD>',
        '</xrds:XRDS>'
    ])
    server_response = urlencode({
        'janrain_nonce': JANRAIN_NONCE,
        'openid.mode': 'id_res',
        'openid.claimed_id': 'http://foobar.livejournal.com/',
        'openid.identity': 'http://foobar.livejournal.com/',
        'openid.op_endpoint': 'http://www.livejournal.com/openid/server.bml',
        'openid.return_to': 'http://myapp.com/complete/livejournal/?'
                            'janrain_nonce=' + JANRAIN_NONCE,
        'openid.response_nonce': JANRAIN_NONCE + 'wGp2rj',
        'openid.assoc_handle': '1364932966:ZTiur8sem3r2jzZougMZ:4d1cc3b44e',
        'openid.ns': 'http://specs.openid.net/auth/2.0',
        'openid.signed': 'mode,claimed_id,identity,op_endpoint,return_to,'
                         'response_nonce,assoc_handle',
        'openid.sig': 'Z8MOozVPTOBhHG5ZS1NeGofxs1Q=',
    })
    server_bml_body = '\n'.join([
        'assoc_handle:1364935340:ZhruPQ7DJ9eGgUkeUA9A:27f8c32464',
        'assoc_type:HMAC-SHA1',
        'dh_server_public:WzsRyLomvAV3vwvGUrfzXDgfqnTF+m1l3JWb55fyHO7visPT4tmQ'
        'iTjqFFnSVAtAOvQzoViMiZQisxNwnqSK4lYexoez1z6pP5ry3pqxJAEYj60vFGvRztict'
        'Eo0brjhmO1SNfjK1ppjOymdykqLpZeaL5fsuLtMCwTnR/JQZVA=',
        'enc_mac_key:LiOEVlLJSVUqfNvb5zPd76nEfvc=',
        'expires_in:1207060',
        'ns:http://specs.openid.net/auth/2.0',
        'session_type:DH-SHA1',
        ''
    ])

    def openid_url(self):
        return super().openid_url() + '/data/yadis'

    def post_start(self):
        self.strategy.remove_from_request_data('openid_lj_user')

    def _setup_handlers(self):
        HTTPretty.register_uri(
            HTTPretty.POST,
            'http://www.livejournal.com/openid/server.bml',
            headers={'Accept-Encoding': 'identity',
                     'Content-Type': 'application/x-www-form-urlencoded'},
            status=200,
            body=self.server_bml_body
        )
        HTTPretty.register_uri(
            HTTPretty.GET,
            'http://foobar.livejournal.com/',
            headers={
                'Accept-Encoding': 'identity',
                'Accept': 'text/html; q=0.3,'
                          'application/xhtml+xml; q=0.5,'
                          'application/xrds+xml'
            },
            status=200,
            body=self.discovery_body
        )

    def test_login(self):
        self.strategy.set_request_data({'openid_lj_user': 'foobar'},
                                       self.backend)
        self._setup_handlers()
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_request_data({'openid_lj_user': 'foobar'},
                                       self.backend)
        self._setup_handlers()
        self.do_partial_pipeline()

    def test_failed_login(self):
        self._setup_handlers()
        with self.assertRaises(AuthMissingParameter):
            self.do_login()



# -*- coding: utf-8 -*-
from ...exceptions import AuthTokenError
import os
import sys
import json
import datetime
import base64
from calendar import timegm

from jose import jwt
from httpretty import HTTPretty


sys.path.insert(0, '..')


TEST_ROOT = os.path.dirname(os.path.dirname(__file__))

JWK_KEY = {
    'kty': 'RSA',
    'd': 'ZmswNokEvBcxW_Kvcy8mWUQOQCBdGbnM0xR7nhvGHC-Q24z3XAQWlMWbsmGc_R1o'
         '_F3zK7DBlc3BokdRaO1KJirNmnHCw5TlnBlJrXiWpFBtVglUg98-4sRRO0VWnGXK'
         'JPOkBQ6b_DYRO3b0o8CSpWowpiV6HB71cjXTqKPZf-aXU9WjCCAtxVjfIxgQFu5I'
         '-G1Qah8mZeY8HK_y99L4f0siZcbUoaIcfeWBhxi14ODyuSAHt0sNEkhiIVBZE7QZ'
         'm-SEP1ryT9VAaljbwHHPmg7NC26vtLZhvaBGbTTJnEH0ZubbN2PMzsfeNyoCIHy4'
         '4QDSpQDCHfgcGOlHY_t5gQ',
    'e': 'AQAB',
    'use': 'sig',
    'kid': 'testkey',
    'alg': 'RS256',
    'n': 'pUfcJ8WFrVue98Ygzb6KEQXHBzi8HavCu8VENB2As943--bHPcQ-nScXnrRFAUg8'
         'H5ZltuOcHWvsGw_AQifSLmOCSWJAPkdNb0w0QzY7Re8NrPjCsP58Tytp5LicF0Ao'
         'Ag28UK3JioY9hXHGvdZsWR1Rp3I-Z3nRBP6HyO18pEgcZ91c9aAzsqu80An9X4DA'
         'b1lExtZorvcd5yTBzZgr-MUeytVRni2lDNEpa6OFuopHXmg27Hn3oWAaQlbymd4g'
         'ifc01oahcwl3ze2tMK6gJxa_TdCf1y99Yq6oilmVvZJ8kwWWnbPE-oDmOVPVnEyT'
         'vYVCvN4rBT1DQ-x0F1mo2Q',
}

JWK_PUBLIC_KEY = {key: value for key, value in JWK_KEY.items() if key != 'd'}


class OpenIdConnectTestMixin:
    """
    Mixin to test OpenID Connect consumers. Inheriting classes should also
    inherit OAuth2Test.
    """
    client_key = 'a-key'
    client_secret = 'a-secret-key'
    issuer = None  # id_token issuer
    openid_config_body = None
    key = None

    def setUp(self):
        super().setUp()
        self.key = JWK_KEY.copy()
        self.public_key = JWK_PUBLIC_KEY.copy()

        HTTPretty.register_uri(HTTPretty.GET,
                               self.backend.OIDC_ENDPOINT + '/.well-known/openid-configuration',
                               status=200,
                               body=self.openid_config_body
                               )
        oidc_config = json.loads(self.openid_config_body)

        def jwks(_request, _uri, headers):
            return 200, headers, json.dumps({'keys': [self.key]})

        HTTPretty.register_uri(HTTPretty.GET,
                               oidc_config.get('jwks_uri'),
                               status=200,
                               body=json.dumps({'keys': [self.public_key]}))

    def extra_settings(self):
        settings = super().extra_settings()
        settings.update({
            'SOCIAL_AUTH_{0}_KEY'.format(self.name): self.client_key,
            'SOCIAL_AUTH_{0}_SECRET'.format(self.name): self.client_secret,
            'SOCIAL_AUTH_{0}_ID_TOKEN_DECRYPTION_KEY'.format(self.name):
                self.client_secret
        })
        return settings

    def get_id_token(self, client_key=None, expiration_datetime=None,
                     issue_datetime=None, nonce=None, issuer=None):
        """
        Return the id_token to be added to the access token body.
        """
        return {
            'iss': issuer,
            'nonce': nonce,
            'aud': client_key,
            'azp': client_key,
            'exp': expiration_datetime,
            'iat': issue_datetime,
            'sub': '1234'
        }

    def prepare_access_token_body(self, client_key=None, tamper_message=False,
                                  expiration_datetime=None,
                                  issue_datetime=None, nonce=None,
                                  issuer=None):
        """
        Prepares a provider access token response. Arguments:

        client_id       -- (str) OAuth ID for the client that requested
                                 authentication.
        expiration_time -- (datetime) Date and time after which the response
                                      should be considered invalid.
        """

        body = {'access_token': 'foobar', 'token_type': 'bearer'}
        client_key = client_key or self.client_key
        now = datetime.datetime.utcnow()
        expiration_datetime = expiration_datetime or \
            (now + datetime.timedelta(seconds=30))
        issue_datetime = issue_datetime or now
        nonce = nonce or 'a-nonce'
        issuer = issuer or self.issuer
        id_token = self.get_id_token(
            client_key,
            timegm(expiration_datetime.utctimetuple()),
            timegm(issue_datetime.utctimetuple()),
            nonce,
            issuer
        )

        body['id_token'] = jwt.encode(
            id_token,
            key=dict(self.key,
                     iat=timegm(issue_datetime.utctimetuple()),
                     nonce=nonce),
            algorithm='RS256',
            access_token='foobar'
        )

        if tamper_message:
            header, msg, sig = body['id_token'].split('.')
            id_token['sub'] = '1235'
            msg = base64.encodebytes(json.dumps(id_token).encode()).decode()
            body['id_token'] = '.'.join([header, msg, sig])

        return json.dumps(body)

    def authtoken_raised(self, expected_message, **access_token_kwargs):
        self.access_token_body = self.prepare_access_token_body(
            **access_token_kwargs
        )
        with self.assertRaisesRegex(AuthTokenError, expected_message):
            self.do_login()

    def test_invalid_signature(self):
        self.authtoken_raised(
            'Token error: Signature verification failed',
            tamper_message=True
        )

    def test_expired_signature(self):
        expiration_datetime = datetime.datetime.utcnow() - \
            datetime.timedelta(seconds=30)
        self.authtoken_raised('Token error: Signature has expired',
                              expiration_datetime=expiration_datetime)

    def test_invalid_issuer(self):
        self.authtoken_raised('Token error: Invalid issuer',
                              issuer='someone-else')

    def test_invalid_audience(self):
        self.authtoken_raised('Token error: Invalid audience',
                              client_key='someone-else')

    def test_invalid_issue_time(self):
        expiration_datetime = datetime.datetime.utcnow() - \
            datetime.timedelta(hours=1)
        self.authtoken_raised('Token error: Incorrect id_token: iat',
                              issue_datetime=expiration_datetime)

    def test_invalid_nonce(self):
        self.authtoken_raised(
            'Token error: Incorrect id_token: nonce',
            nonce='something-wrong'
        )



"""
Stackoverflow OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/stackoverflow.html
"""
from .oauth import BaseOAuth2


class StackoverflowOAuth2(BaseOAuth2):
    """Stackoverflow OAuth2 authentication backend"""
    name = 'stackoverflow'
    ID_KEY = 'user_id'
    AUTHORIZATION_URL = 'https://stackexchange.com/oauth'
    ACCESS_TOKEN_URL = 'https://stackexchange.com/oauth/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ','
    EXTRA_DATA = [
        ('id', 'id'),
        ('expires', 'expires')
    ]

    def get_user_details(self, response):
        """Return user details from Stackoverflow account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('display_name')
        )
        return {'username': response.get('link').rsplit('/', 1)[-1],
                'full_name': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://api.stackexchange.com/2.1/me',
            params={
                'site': 'stackoverflow',
                'access_token': access_token,
                'key': self.setting('API_KEY')
            }
        )['items'][0]

    def request_access_token(self, *args, **kwargs):
        return self.get_querystring(*args, **kwargs)



from urllib.parse import urljoin

from social_core.utils import cache

from ..utils import append_slash
from .open_id_connect import OpenIdConnectAuth


class Fence(OpenIdConnectAuth):

    name = 'fence'
    OIDC_ENDPOINT = 'https://nci-crdc.datacommons.io'
    ID_KEY = 'username'
    ACCESS_TOKEN_METHOD = 'POST'
    DEFAULT_SCOPE = ['openid', 'user']
    JWT_DECODE_OPTIONS = {'verify_at_hash': False}

    def _url(self, path):
        return urljoin(append_slash(self.OIDC_ENDPOINT), path)

    def authorization_url(self):
        return self._url('user/oauth2/authorize')

    def access_token_url(self):
        return self._url('user/oauth2/token')

    @cache(ttl=86400)
    def oidc_config(self):
        return self.get_json(self._url('.well-known/openid-configuration'))

    def get_user_details(self, response):
        return {
            'username': response.get('preferred_username'),
            'email': response.get('username'),
            'fullname': response.get('name'),
            'first_name': response.get('given_name'),
            'last_name': response.get('family_name'),
        }



import unittest

from ...backends.base import BaseAuth
from ..strategy import TestStrategy
from ..models import TestStorage


class BrokenBackendAuth(BaseAuth):
    name = 'broken'


class BrokenBackendTest(unittest.TestCase):
    def setUp(self):
        self.backend = BrokenBackendAuth(TestStrategy(TestStorage))

    def tearDown(self):
        self.backend = None

    def test_auth_url(self):
        with self.assertRaisesRegex(NotImplementedError,
                                    'Implement in subclass'):
            self.backend.auth_url()

    def test_auth_html(self):
        with self.assertRaisesRegex(NotImplementedError,
                                    'Implement in subclass'):
            self.backend.auth_html()

    def test_auth_complete(self):
        with self.assertRaisesRegex(NotImplementedError,
                                    'Implement in subclass'):
            self.backend.auth_complete()

    def test_get_user_details(self):
        with self.assertRaisesRegex(NotImplementedError,
                                    'Implement in subclass'):
            self.backend.get_user_details(None)



import json

from .oauth import OAuth2Test


class NaverOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.naver.NaverOAuth2'
    user_data_url = 'https://openapi.naver.com/v1/nid/me'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
    })

    user_data_content_type = 'text/json'
    user_data_body = json.dumps({
        'resultcode': '00',
        'message': 'success',
        'response': {
            'email': 'openapi@naver.com',
            'nickname': 'foobar',
            'profile_image': 'https://ssl.pstatic.net/static/pwe/address/nodata_33x33.gif',
            'age': '40-49',
            'gender': 'F',
            'id': '32742776',
            'name': 'foobar',
            'birthday': '10-01',
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import base64

from .oauth import BaseOAuth2


class ZoomOAuth2(BaseOAuth2):
    """
    Zoom OAuth2 authentication backend
    Doc Reference: https://marketplace.zoom.us/docs/guides/auth/oauth
    """
    name = 'zoom-oauth2'
    AUTHORIZATION_URL = 'https://zoom.us/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://zoom.us/oauth/token'
    USER_DETAILS_URL = 'https://api.zoom.us/v2/users/me'
    DEFAULT_SCOPE = ['user:read']
    ACCESS_TOKEN_METHOD = 'POST'
    REFRESH_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('expires_in', 'expires')
    ]

    def user_data(self, access_token, *args, **kwargs):
        response = self.get_json(
            self.USER_DETAILS_URL, headers={
                'Authorization': 'Bearer {access_token}'.format(
                    access_token=access_token
                )
            }
        )
        return response

    def get_user_details(self, response):
        username = response.get('id', '')
        first_name = response.get('first_name', '')
        last_name = response.get('last_name', '')
        email = response.get('email', '')
        fullname = ''
        return {
            'username': username,
            'email': email,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
        }

    def auth_complete_params(self, state=None):
        return {
            'grant_type': 'authorization_code',  # request auth code
            'code': self.data.get('code', ''),  # server response code
            'redirect_uri': self.get_redirect_uri(state),
        }

    def auth_headers(self):
        return {
            'Authorization': b'Basic ' + base64.urlsafe_b64encode(
                '{0}:{1}'.format(*self.get_key_and_secret()).encode()
            )
        }

    def refresh_token_params(self, token, *args, **kwargs):
        return {'refresh_token': token, 'grant_type': 'refresh_token'}



import base64

from .oauth import BaseOAuth2


class PayPalOAuth2(BaseOAuth2):
    """
    PayPal OAuth2 backend, docs at:
        https://developer.paypal.com/docs/connect-with-paypal/integrate/
    """

    name = 'paypal-oauth2'
    ID_KEY = 'user_id'
    AUTHORIZATION_URL = 'https://www.paypal.com/connect'
    ACCESS_TOKEN_URL = 'https://api.paypal.com/v1/oauth2/token'
    USER_DATA_URL = (
        'https://api.paypal.com/v1/identity/oauth2/userinfo?schema=paypalv1.1'
    )
    DEFAULT_SCOPE = ['openid', 'profile']
    ACCESS_TOKEN_METHOD = 'POST'
    REFRESH_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def user_data(self, access_token, *args, **kwargs):
        auth_header = {'Authorization': 'Bearer %s' % access_token}
        response = self.get_json(self.USER_DATA_URL, headers=auth_header)
        return response

    def get_user_details(self, response):
        username = response.get(self.ID_KEY).split('/')[-1]
        fullname, first_name, last_name = self.get_user_names(
            response.get('name', ''),
            response.get('given_name', ''),
            response.get('family_name', ''),
        )
        emails = response.get('emails', [])
        email = self.get_email(emails)
        return {
            'username': username,
            'email': email,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
        }

    def auth_complete_params(self, state=None):
        return {
            'grant_type': 'authorization_code',
            'code': self.data.get('code', ''),
        }

    def auth_headers(self):
        auth = ('%s:%s' % self.get_key_and_secret()).encode()
        return {'Authorization': b'Basic ' + base64.urlsafe_b64encode(auth)}

    def refresh_token_params(self, token, *args, **kwargs):
        return {'refresh_token': token, 'grant_type': 'refresh_token'}

    @staticmethod
    def get_email(emails):
        if not emails:
            return ''
        primary_emails = (email for email in emails
                          if email.get('primary', False))
        primary_or_first = next(primary_emails, emails[0])
        return primary_or_first.get('value')


class PayPalOAuth2Sandbox(PayPalOAuth2):
    name = 'paypal-oauth2-sandbox'
    AUTHORIZATION_URL = 'https://www.sandbox.paypal.com/connect'
    ACCESS_TOKEN_URL = 'https://api.sandbox.paypal.com/v1/oauth2/token'
    USER_DATA_URL = (
        'https://api.sandbox.paypal.com/v1/identity/oauth2/userinfo?schema=paypalv1.1'
    )



import json

from .oauth import OAuth2Test


class StravaOAuthTest(OAuth2Test):
    backend_path = 'social_core.backends.strava.StravaOAuth'
    user_data_url = 'https://www.strava.com/api/v3/athlete'
    expected_username = 'marianne_v'
    access_token_body = json.dumps({
        'token_type': 'Bearer',
        'expires_at': 1572805000,
        'expires_in': 227615,
        'refresh_token': 'f51defab4632d27255dd0d106504dfd7568fd1df6',
        'access_token': '83ebeabdec09f6670863766f792ead24d61fe3f9',
        'athlete': {
            'id': 1234567890987654321,
            'username': 'marianne_v',
            'resource_state': 2,
            'firstname': 'Marianne',
            'lastname': 'V.',
            'city': 'Francisco',
            'state': 'California',
            'country': 'United States',
            'sex': 'F',
            'premium': 'true',
            'summit': 'true',
            'created_at': '2017-11-14T02:30:05Z',
            'updated_at': '2018-02-06T19:32:20Z',
            'badge_type_id': 4,
            'profile_medium': 'https://xxxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/medium.jpg',
            'profile': 'https://xxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/large.jpg',
            'friend': 'null',
            'follower': 'null'
        }
    })
    user_data_body = json.dumps({
        'id': 1234567890987654321,
        'username': 'marianne_v',
        'resource_state': 3,
        'firstname': 'Marianne',
        'lastname': 'V.',
        'city': 'San Francisco',
        'state': 'CA',
        'country': 'US',
        'sex': 'F',
        'premium': 'true',
        'created_at': '2017-11-14T02:30:05Z',
        'updated_at': '2018-02-06T19:32:20Z',
        'badge_type_id': 4,
        'profile_medium': 'https://xxxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/medium.jpg',
        'profile': 'https://xxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/large.jpg',
        'friend': 'null',
        'follower': 'null',
        'follower_count': 5,
        'friend_count': 5,
        'mutual_friend_count': 0,
        'athlete_type': 1,
        'date_preference': '%m/%d/%Y',
        'measurement_preference': 'feet',
        'clubs': [],
        'ftp': 'null',
        'weight': 0,
        'bikes': [],
        'shoes': []
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test

from unittest.mock import patch


TEST_KEY = """
-----BEGIN EC PRIVATE KEY-----
MHcCAQEEIKQya8aIoeoOLeThk7Ad/lLyAo2fTp9IuhIpy2CivH/qoAoGCCqGSM49
AwEHoUQDQgAEyEY7IMlNJtyaF/pdcM/PpQ8OCe19Sf1Yxq4HQsrB2b7QogB95Vjt
6mTZDAhlXIBtuM/JLrdkMfPmwjVKLgxHAQ==
-----END EC PRIVATE KEY-----
"""


token_data = {
    'sub': '11011110101011011011111011101111',
    'first_name': 'Foo',
    'last_name': 'Bar',
    'email': 'foobar@apple.com',
}


class AppleIdTest(OAuth2Test):
    backend_path = 'social_core.backends.apple.AppleIdAuth'
    user_data_url = 'https://appleid.apple.com/auth/authorize/'
    id_token = 'a-id-token'
    access_token_body = json.dumps({'id_token': id_token, 'access_token': 'a-test-token'})
    expected_username = token_data['sub']

    def extra_settings(self):
        return {
            'SOCIAL_AUTH_' + self.name + '_TEAM': 'a-team-id',
            'SOCIAL_AUTH_' + self.name + '_KEY': 'a-key-id',
            'SOCIAL_AUTH_' + self.name + '_CLIENT': 'a-client-id',
            'SOCIAL_AUTH_' + self.name + '_SECRET': TEST_KEY,
            'SOCIAL_AUTH_' + self.name + '_SCOPE': ['name', 'email'],
        }

    def test_login(self):
        with patch('{}.{}'.format(self.backend_path, 'decode_id_token'),
                   return_value=token_data) as decode_mock:
            self.do_login()
        assert decode_mock.called
        assert decode_mock.call_args[0] == (self.id_token,)

    def test_partial_pipeline(self):
        with patch('{}.{}'.format(self.backend_path, 'decode_id_token'),
                   return_value=token_data) as decode_mock:
            self.do_partial_pipeline()
        assert decode_mock.called
        assert decode_mock.call_args[0] == (self.id_token,)



from .oauth import BaseOAuth2


class NaverOAuth2(BaseOAuth2):
    """Naver OAuth authentication backend"""
    name = 'naver'
    AUTHORIZATION_URL = 'https://nid.naver.com/oauth2.0/authorize'
    ACCESS_TOKEN_URL = 'https://nid.naver.com/oauth2.0/token'
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [
        ('id', 'id'),
    ]

    def get_user_id(self, details, response):
        return response.get('id')

    def get_user_details(self, response):
        """Return user details from Naver account"""
        return {
            'username': response.get('username'),
            'email': response.get('email'),
            'fullname': response.get('username'),
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        response = self.request(
            'https://openapi.naver.com/v1/nid/me',
            headers={
                'Authorization': 'Bearer {0}'.format(access_token),
                'Content_Type': 'text/json'
            }
        )

        data = response.json()

        return {
            'id': self._fetch(data, 'id'),
            'email': self._fetch(data, 'email'),
            'username': self._fetch(data, 'name'),
            'nickname': self._fetch(data, 'nickname'),
            'gender': self._fetch(data, 'gender'),
            'age': self._fetch(data, 'age'),
            'birthday': self._fetch(data, 'birthday'),
            'profile_image': self._fetch(data, 'profile_image')
        }

    def auth_headers(self):
        client_id, client_secret = self.get_key_and_secret()
        return {
            'grant_type': 'authorization_code',
            'code': self.data.get('code'),
            'client_id': client_id,
            'client_secret': client_secret,
        }

    def _fetch(self, data, key):
        try:
            return data['response'][key]
        except (KeyError, TypeError):
            return ''



SERIALIZABLE_TYPES = (dict, list, tuple, set, bool, type(None), int, str, bytes)


def is_dict_type(value):
    """Treat any dict, MergeDict, MultiDict instance as dict type"""
    # Check by class name to avoid importing Django MergeDict or
    # Werkzeug MultiDict
    return isinstance(value, dict) or \
        value.__class__.__name__ in ('MergeDict', 'MultiDict')


def partial_prepare(strategy, backend, next_step, user=None, social=None,
                    *args, **kwargs):
    kwargs.update({
        'response': kwargs.get('response') or {},
        'details': kwargs.get('details') or {},
        'username': kwargs.get('username'),
        'uid': kwargs.get('uid'),
        'is_new': kwargs.get('is_new') or False,
        'new_association': kwargs.get('new_association') or False,
        'user': hasattr(user, 'id') and user.id or None,
        'social': social and {
            'provider': social.provider,
            'uid': social.uid
        } or None
    })

    clean_args = [strategy.to_session_value(val) for val in args]

    # Clean any MergeDict data type from the values
    clean_kwargs = {}
    for name, value in kwargs.items():
        value = dict(value) if is_dict_type(value) else value
        if isinstance(value, SERIALIZABLE_TYPES):
            clean_kwargs[name] = strategy.to_session_value(value)

    return strategy.storage.partial.prepare(backend.name, next_step, {
        'args': clean_args,
        'kwargs': clean_kwargs
    })


def partial_store(strategy, backend, next_step, user=None, social=None,
                  *args, **kwargs):
    partial = partial_prepare(strategy, backend, next_step, user=user,
                              social=social, *args, **kwargs)
    return strategy.storage.partial.store(partial)


def partial_load(strategy, token):
    partial = strategy.storage.partial.load(token)

    if partial:
        args = partial.args
        kwargs = partial.kwargs.copy()
        user = kwargs.get('user')
        social = kwargs.get('social')

        if isinstance(social, dict):
            kwargs['social'] = strategy.storage.user.get_social_auth(**social)

        if user:
            kwargs['user'] = strategy.storage.user.get_user(user)

        partial.args = [strategy.from_session_value(val) for val in args]
        partial.kwargs = dict((key, strategy.from_session_value(val))
                              for key, val in kwargs.items())
    return partial



from urllib.parse import quote

from .utils import sanitize_redirect, user_is_authenticated, \
                   user_is_active, partial_pipeline_data, setting_url


def do_auth(backend, redirect_name='next'):
    # Save any defined next value into session
    data = backend.strategy.request_data(merge=False)

    # Save extra data into session.
    for field_name in backend.setting('FIELDS_STORED_IN_SESSION', []):
        if field_name in data:
            backend.strategy.session_set(field_name, data[field_name])
        else:
            backend.strategy.session_set(field_name, None)

    if redirect_name in data:
        # Check and sanitize a user-defined GET/POST next field value
        redirect_uri = data[redirect_name]
        if backend.setting('SANITIZE_REDIRECTS', True):
            allowed_hosts = backend.setting('ALLOWED_REDIRECT_HOSTS', []) + \
                            [backend.strategy.request_host()]
            redirect_uri = sanitize_redirect(allowed_hosts, redirect_uri)
        backend.strategy.session_set(
            redirect_name,
            redirect_uri or backend.setting('LOGIN_REDIRECT_URL')
        )
    return backend.start()


def do_complete(backend, login, user=None, redirect_name='next',
                *args, **kwargs):
    data = backend.strategy.request_data()

    is_authenticated = user_is_authenticated(user)
    user = user if is_authenticated else None

    partial = partial_pipeline_data(backend, user, *args, **kwargs)
    if partial:
        user = backend.continue_pipeline(partial)
        # clean partial data after usage
        backend.strategy.clean_partial_pipeline(partial.token)
    else:
        user = backend.complete(user=user, *args, **kwargs)

    # pop redirect value before the session is trashed on login(), but after
    # the pipeline so that the pipeline can change the redirect if needed
    redirect_value = backend.strategy.session_get(redirect_name, '') or \
        data.get(redirect_name, '')

    # check if the output value is something else than a user and just
    # return it to the client
    user_model = backend.strategy.storage.user.user_model()
    if user and not isinstance(user, user_model):
        return user

    if is_authenticated:
        if not user:
            url = setting_url(backend, redirect_value, 'LOGIN_REDIRECT_URL')
        else:
            url = setting_url(backend, redirect_value,
                              'NEW_ASSOCIATION_REDIRECT_URL',
                              'LOGIN_REDIRECT_URL')
    elif user:
        if user_is_active(user):
            # catch is_new/social_user in case login() resets the instance
            is_new = getattr(user, 'is_new', False)
            social_user = user.social_user
            login(backend, user, social_user)
            # store last login backend name in session
            backend.strategy.session_set('social_auth_last_login_backend',
                                         social_user.provider)

            if is_new:
                url = setting_url(backend,
                                  'NEW_USER_REDIRECT_URL',
                                  redirect_value,
                                  'LOGIN_REDIRECT_URL')
            else:
                url = setting_url(backend, redirect_value,
                                  'LOGIN_REDIRECT_URL')
        else:
            if backend.setting('INACTIVE_USER_LOGIN', False):
                social_user = user.social_user
                login(backend, user, social_user)
            url = setting_url(backend, 'INACTIVE_USER_URL', 'LOGIN_ERROR_URL',
                              'LOGIN_URL')
    else:
        url = setting_url(backend, 'LOGIN_ERROR_URL', 'LOGIN_URL')

    if redirect_value and redirect_value != url:
        redirect_value = quote(redirect_value)
        url += ('&' if '?' in url else '?') + \
            '{0}={1}'.format(redirect_name, redirect_value)

    if backend.setting('SANITIZE_REDIRECTS', True):
        allowed_hosts = backend.setting('ALLOWED_REDIRECT_HOSTS', []) + \
                        [backend.strategy.request_host()]
        url = sanitize_redirect(allowed_hosts, url) or \
            backend.setting('LOGIN_REDIRECT_URL')
    return backend.strategy.redirect(url)


def do_disconnect(backend, user, association_id=None, redirect_name='next',
                  *args, **kwargs):
    partial = partial_pipeline_data(backend, user, *args, **kwargs)
    if partial:
        if association_id and not partial.kwargs.get('association_id'):
            partial.extend_kwargs({
                'association_id': association_id
            })
        response = backend.disconnect(*partial.args, **partial.kwargs)
        # clean partial data after usage
        backend.strategy.clean_partial_pipeline(partial.token)
    else:
        response = backend.disconnect(user=user, association_id=association_id,
                                      *args, **kwargs)

    if isinstance(response, dict):
        url = backend.strategy.absolute_uri(
            backend.strategy.request_data().get(redirect_name, '') or
            backend.setting('DISCONNECT_REDIRECT_URL') or
            backend.setting('LOGIN_REDIRECT_URL')
        )
        if backend.setting('SANITIZE_REDIRECTS', True):
            allowed_hosts = backend.setting('ALLOWED_REDIRECT_HOSTS', []) + \
                            [backend.strategy.request_host()]
            url = sanitize_redirect(allowed_hosts, url) or \
                backend.setting('DISCONNECT_REDIRECT_URL') or \
                backend.setting('LOGIN_REDIRECT_URL')
        response = backend.strategy.redirect(url)
    return response



import json

from .oauth import OAuth2Test


class PatreonOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.patreon.PatreonOAuth2'
    user_data_url = 'https://www.patreon.com/api/oauth2/v2/identity?' + \
        'fields%5Buser%5D=about,created,email,first_name,full_name,' + \
        'image_url,last_name,social_connections,thumb_url,url,vanity'
    expected_username = 'JohnInterwebs'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
    })
    user_data_body = json.dumps({
        'data': {
            'relationships': {
                'pledges': {
                    'data': [{
                        'type': 'pledge', 'id': '123456'
                    }]
                }
            },
            'attributes': {
                'last_name': 'Interwebs',
                'is_suspended': False,
                'has_password': True,
                'full_name': 'John Interwebs',
                'is_nuked': False,
                'first_name': 'John',
                'social_connections': {
                    'spotify': None,
                    'discord': None,
                    'twitter': None,
                    'youtube': None,
                    'facebook': None,
                    'deviantart': None,
                    'twitch': None
                },
                'twitter': None,
                'is_email_verified': True,
                'facebook_id': None,
                'email': 'john@example.com',
                'facebook': None,
                'thumb_url': 'https://c8.patreon.com/100/123456',
                'vanity': None,
                'about': None,
                'is_deleted': False,
                'created': '2017-05-05T05:16:34+00:00',
                'url': 'https://www.patreon.com/user?u=123456',
                'gender': 0,
                'youtube': None,
                'discord_id': None,
                'image_url': 'https://c8.patreon.com/400/123456',
                'twitch': None
            },
            'type': 'user',
            'id': '123456'
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class MRGOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.mailru.MRGOAuth2'
    user_data_url = 'https://oauth.mail.ru/userinfo'
    expected_username = 'FooBar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'first_name': 'Foo',
        'last_name': 'Bar',
        'name': 'Foo Bar',
        'locale': 'ru_RU',
        'email': 'foobar@example.com',
        'birthday': '11.07.1970',
        'gender': 'm',
        'image': 'http://cs7003.vk.me/v7003815/22a1/xgG9fb-IJ3Y.jpg',

    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from collections import OrderedDict

from .base import BaseAuth
from ..exceptions import MissingBackend
from ..utils import module_member, user_is_authenticated


# Cache for discovered backends.
BACKENDSCACHE = OrderedDict()


def load_backends(backends, force_load=False):
    """
    Load backends defined on SOCIAL_AUTH_AUTHENTICATION_BACKENDS, backends will
    be imported and cached on BACKENDSCACHE. The key in that dict will be the
    backend name, and the value is the backend class.

    Only subclasses of BaseAuth (and sub-classes) are considered backends.

    Previously there was a BACKENDS attribute expected on backends modules,
    this is not needed anymore since it's enough with the
    AUTHENTICATION_BACKENDS setting. BACKENDS was used because backends used to
    be split on two classes the authentication backend and another class that
    dealt with the auth mechanism with the provider, those classes are joined
    now.

    A force_load boolean argument is also provided so that get_backend
    below can retry a requested backend that may not yet be discovered.
    """
    global BACKENDSCACHE
    if force_load:
        BACKENDSCACHE = OrderedDict()
    if not BACKENDSCACHE:
        for auth_backend in backends:
            backend = module_member(auth_backend)
            if issubclass(backend, BaseAuth):
                BACKENDSCACHE[backend.name] = backend
    return BACKENDSCACHE


def get_backend(backends, name):
    """Returns a backend by name. Backends are stored in the BACKENDSCACHE
    cache dict. If not found, each of the modules referenced in
    AUTHENTICATION_BACKENDS is imported and checked for a BACKENDS
    definition. If the named backend is found in the module's BACKENDS
    definition, it's then stored in the cache for future access.
    """
    try:
        # Cached backend which has previously been discovered
        return BACKENDSCACHE[name]
    except KeyError:
        # Reload BACKENDS to ensure a missing backend hasn't been missed
        load_backends(backends, force_load=True)
        try:
            return BACKENDSCACHE[name]
        except KeyError:
            raise MissingBackend(name)


def user_backends_data(user, backends, storage):
    """
    Will return backends data for given user, the return value will have the
    following keys:
        associated: UserSocialAuth model instances for currently associated
                    accounts
        not_associated: Not associated (yet) backend names
        backends: All backend names.

    If user is not authenticated, then 'associated' list is empty, and there's
    no difference between 'not_associated' and 'backends'.
    """
    available = list(load_backends(backends).keys())
    values = {'associated': [],
              'not_associated': available,
              'backends': available}
    if user_is_authenticated(user):
        associated = storage.user.get_social_auth_for_user(user)
        not_associated = list(set(available) -
                              set(assoc.provider for assoc in associated))
        values['associated'] = associated
        values['not_associated'] = not_associated
    return values



"""
LINE Login OAuth2 backend, docs at:
    https://developers.line.me/en/docs/line-login/
"""
import requests
import json

from .oauth import BaseOAuth2
from ..exceptions import AuthFailed
from ..utils import handle_http_errors


class LineOAuth2(BaseOAuth2):
    name = 'line'
    AUTHORIZATION_URL = 'https://access.line.me/oauth2/v2.1/authorize'
    ACCESS_TOKEN_URL = 'https://api.line.me/oauth2/v2.1/token'
    BASE_API_URL = 'https://api.line.me'
    USER_INFO_URL = BASE_API_URL + '/v2/profile'
    ACCESS_TOKEN_METHOD = 'POST'
    STATE_PARAMETER = True
    DEFAULT_SCOPE = ['profile']
    REDIRECT_STATE = True
    ID_KEY = 'userId'
    EXTRA_DATA = [
        ('userId', 'id'),
        ('picture_url', 'picture_url'),
        ('status_message', 'status_message'),
        ('expires_in', 'expire'),
        ('refresh_token', 'refresh_token')
    ]

    def auth_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            'response_type': self.RESPONSE_TYPE,
            'client_id': client_id,
            'redirect_uri': self.get_redirect_uri(),
            'state': self.get_or_create_state(),
            'scope': self.get_scope()
        }

    def process_error(self, data):
        error_code = data.get('errorCode') or \
            data.get('statusCode') or \
            data.get('error')
        error_message = data.get('errorMessage') or \
            data.get('error_description')
        if error_code is not None or error_message is not None:
            raise AuthFailed(self, error_message or error_code)

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)

        try:
            response = self.request_access_token(
                self.access_token_url(),
                method=self.ACCESS_TOKEN_METHOD,
                headers=self.auth_headers(),
                data=self.auth_complete_params()
            )
            self.process_error(response)

            return self.do_auth(response['access_token'], response=response,
                                *args, **kwargs)
        except requests.HTTPError as err:
            self.process_error(json.loads(err.response.content))

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response.get('displayName')
        )
        username = response.get('userId')
        picture_url = response.get('pictureUrl')
        status_message = response.get('statusMessage')
        return {
            'username': username,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
            'picture_url': picture_url,
            'status_message': status_message
        }

    def get_user_id(self, details, response):
        """
        Return a unique ID for the current user, by default from
        server response.
        """
        return response.get(self.ID_KEY)

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        try:
            response = self.get_json(
                self.USER_INFO_URL,
                headers={
                    'Authorization': 'Bearer {}'.format(access_token)
                }
            )
            self.process_error(response)
            return response
        except requests.HTTPError as err:
            self.process_error(err.response.json())



"""
Odnoklassniki OAuth2 and Iframe Application backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/odnoklassnikiru.html
"""
from hashlib import md5
from urllib.parse import unquote

from .base import BaseAuth
from .oauth import BaseOAuth2
from ..exceptions import AuthFailed


class OdnoklassnikiOAuth2(BaseOAuth2):
    """Odnoklassniki authentication backend"""
    name = 'odnoklassniki-oauth2'
    ID_KEY = 'uid'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ';'
    AUTHORIZATION_URL = 'https://connect.ok.ru/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://api.ok.ru/oauth/token.do'
    EXTRA_DATA = [('refresh_token', 'refresh_token'),
                  ('expires_in', 'expires')]

    def get_user_details(self, response):
        """Return user details from Odnoklassniki request"""
        fullname, first_name, last_name = self.get_user_names(
            fullname=unquote(response['name']),
            first_name=unquote(response['first_name']),
            last_name=unquote(response['last_name'])
        )
        return {
            'username': response['uid'],
            'email': response.get('email', ''),
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data from Odnoklassniki REST API"""
        data = {'access_token': access_token, 'method': 'users.getCurrentUser'}
        key, secret = self.get_key_and_secret()
        public_key = self.setting('PUBLIC_NAME')
        return odnoklassniki_api(self, data, 'https://api.ok.ru/',
                                 public_key, secret, 'oauth')


class OdnoklassnikiApp(BaseAuth):
    """Odnoklassniki iframe app authentication backend"""
    name = 'odnoklassniki-app'
    ID_KEY = 'uid'

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        return dict([(key, value) for key, value in response.items()
                     if key in response['extra_data_list']])

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            fullname=unquote(response['name']),
            first_name=unquote(response['first_name']),
            last_name=unquote(response['last_name'])
        )
        return {
            'username': response['uid'],
            'email': '',
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def auth_complete(self, *args, **kwargs):
        self.verify_auth_sig()
        response = self.get_response()
        fields = ('uid', 'first_name', 'last_name', 'name') + \
            self.setting('EXTRA_USER_DATA_LIST', ())
        data = {
            'method': 'users.getInfo',
            'uids': '{0}'.format(response['logged_user_id']),
            'fields': ','.join(fields),
        }
        client_key, client_secret = self.get_key_and_secret()
        public_key = self.setting('PUBLIC_NAME')
        details = odnoklassniki_api(self, data, response['api_server'],
                                    public_key, client_secret,
                                    'iframe_nosession')
        if len(details) == 1 and 'uid' in details[0]:
            details = details[0]
            auth_data_fields = self.setting('EXTRA_AUTH_DATA_LIST',
                                            ('api_server', 'apiconnection',
                                             'session_key', 'authorized',
                                             'session_secret_key'))

            for field in auth_data_fields:
                details[field] = response[field]
            details['extra_data_list'] = fields + auth_data_fields
            kwargs.update({'backend': self, 'response': details})
        else:
            raise AuthFailed(self, 'Cannot get user details: API error')
        return self.strategy.authenticate(*args, **kwargs)

    def get_auth_sig(self):
        secret_key = self.setting('SECRET')
        hash_source = '{0:s}{1:s}{2:s}'.format(self.data['logged_user_id'],
                                               self.data['session_key'],
                                               secret_key)
        return md5(hash_source.encode('utf-8')).hexdigest()

    def get_response(self):
        fields = ('logged_user_id', 'api_server', 'application_key',
                  'session_key', 'session_secret_key', 'authorized',
                  'apiconnection')
        return dict((name, self.data[name]) for name in fields
                    if name in self.data)

    def verify_auth_sig(self):
        correct_key = self.get_auth_sig()
        key = self.data['auth_sig'].lower()
        if correct_key != key:
            raise AuthFailed(self, 'Wrong authorization key')


def odnoklassniki_oauth_sig(data, client_secret):
    """
    Calculates signature of request data access_token value must be included
    Algorithm is described at
        https://apiok.ru/wiki/pages/viewpage.action?pageId=12878032,
    search for "little bit different way"
    """
    suffix = md5(
        '{0:s}{1:s}'.format(data['access_token'],
                            client_secret).encode('utf-8')
    ).hexdigest()
    check_list = sorted(['{0:s}={1:s}'.format(key, value)
                         for key, value in data.items()
                         if key != 'access_token'])
    return md5((''.join(check_list) + suffix).encode('utf-8')).hexdigest()


def odnoklassniki_iframe_sig(data, client_secret_or_session_secret):
    """
    Calculates signature as described at:
        https://apiok.ru/wiki/display/ok/Authentication+and+Authorization
    If API method requires session context, request is signed with session
    secret key. Otherwise it is signed with application secret key
    """
    param_list = sorted(['{0:s}={1:s}'.format(key, value)
                         for key, value in data.items()])
    return md5(
        (''.join(param_list) + client_secret_or_session_secret).encode('utf-8')
    ).hexdigest()


def odnoklassniki_api(backend, data, api_url, public_key, client_secret,
                      request_type='oauth'):
    """Calls Odnoklassniki REST API method
    https://apiok.ru/wiki/display/ok/Odnoklassniki+Rest+API"""
    data.update({
        'application_key': public_key,
        'format': 'JSON'
    })
    if request_type == 'oauth':
        data['sig'] = odnoklassniki_oauth_sig(data, client_secret)
    elif request_type == 'iframe_session':
        data['sig'] = odnoklassniki_iframe_sig(data,
                                               data['session_secret_key'])
    elif request_type == 'iframe_nosession':
        data['sig'] = odnoklassniki_iframe_sig(data, client_secret)
    else:
        msg = 'Unknown request type {0}. How should it be signed?'
        raise AuthFailed(backend, msg.format(request_type))
    return backend.get_json(api_url + 'fb.do', params=data)



from urllib.parse import urlencode

from .oauth import OAuth1Test


class ZoteroOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.zotero.ZoteroOAuth'
    expected_username = 'FooBar'
    access_token_body = urlencode({
        'oauth_token': 'foobar',
        'oauth_token_secret': 'rodgsNDK4hLJU1504Atk131G',
        'userID': '123456_abcdef',
        'username': 'FooBar'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class ItembaseOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.itembase.ItembaseOAuth2'
    user_data_url = 'https://users.itembase.com/v1/me'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar-token',
        'expires_in': 2592000,
        'token_type': 'bearer',
        'scope': 'user.minimal',
        'refresh_token': 'foobar-refresh-token'
    })
    user_data_body = json.dumps({
        'uuid': 'a4b91ee7-ec1a-49b9-afce-371dc8797749',
        'username': 'foobar',
        'email': 'foobar@itembase.biz',
        'first_name': 'Foo',
        'middle_name': None,
        'last_name': 'Bar',
        'name_format': 'first middle last',
        'locale': 'en',
        'preferred_currency': 'EUR'
    })
    refresh_token_body = json.dumps({
        'access_token': 'foobar-new-token',
        'expires_in': 2592000,
        'token_type': 'bearer',
        'scope': 'user.minimal',
        'refresh_token': 'foobar-new-refresh-token'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class ItembaseOAuth2SandboxTest(OAuth2Test):
    backend_path = 'social_core.backends.itembase.ItembaseOAuth2Sandbox'
    user_data_url = 'http://sandbox.users.itembase.io/v1/me'



import json
from urllib.parse import urlencode

from .oauth import OAuth1Test


class FiveHundredPxOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.five_hundred_px.FiveHundredPxOAuth'
    user_data_url = 'https://api.500px.com/v1/users'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    user_data_body = json.dumps({
        'user': {
            'id': 10101010,
            'username': 'foobar',
            'firstname': '',
            'lastname': '',
            'birthday': None,
            'sex': 1,
            'city': '',
            'state': '',
            'country': '',
            'registration_date': '2011-11-11T00:00:00-00:00',
            'about': '',
            'usertype': 0,
            'domain': 'UserName.500px.com',
            'fotomoto_on': False,
            'locale': 'en',
            'show_nude': False,
            'allow_sale_requests': 1,
            'fullname': 'UserName',
            'userpic_url': 'https://graph.facebook.com/v2.7/'
                           '1000000000/picture?height=100&width=100',
            'userpic_https_url': 'https://graph.facebook.com/v2.7/'
                                 '1000000000/picture?'
                                 'height=100&width=100',
            'cover_url': None,
            'upgrade_status': 0,
            'store_on': False,
            'photos_count': 0,
            'galleries_count': 0,
            'affection': 51,
            'in_favorites_count': 0,
            'friends_count': 2,
            'followers_count': 3,
            'analytics_code': None,
            'invite_pending': False,
            'invite_accepted': False,
            'email': 'user@user.com',
            'shadow_email': 'user@user.com',
            'upload_limit': 20,
            'upload_limit_expiry': '2021-11-11T00:00:00-00:00',
            'upgrade_type': 0,
            'upgrade_status_expiry': '2011-11-21',
            'portfolio_enabled': False,
            'auth': {
                'facebook': 1,
                'twitter': 0,
                'google_oauth2': 1
            },
            'presubmit_for_licensing': None,
            'contacts': {
                'facebook': '1000000000'
            },
            'equipment': {},
            'avatars': {
                'default': {
                    'http': 'https://graph.facebook.com/v2.7/'
                            '1000000000/picture?height=100&width=100',
                    'https': 'https://graph.facebook.com/v2.7/'
                             '1000000000/picture?height=100&width=100'
                },
                'large': {
                    'http': 'https://graph.facebook.com/v2.7/'
                            '1000000000/picture?height=100&width=100',
                    'https': 'https://graph.facebook.com/v2.7/'
                             '1000000000/picture?height=100&width=100'
                },
                'small': {
                    'http': 'https://graph.facebook.com/v2.7/'
                            '1000000000/picture?height=100&width=100',
                    'https': 'https://graph.facebook.com/v2.7/'
                             '1000000000/picture?height=100&width=100'
                },
                'tiny': {
                    'http': 'https://graph.facebook.com/v2.7/'
                            '1000000000/picture?height=100&width=100',
                    'https': 'https://graph.facebook.com/v2.7/'
                             '1000000000/picture?height=100&width=100'
                }
            }
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import unittest as unittest

from ..models import TestStorage
from ..strategy import TestStrategy
from ...backends.utils import load_backends, get_backend
from ...backends.github import GithubOAuth2
from ...exceptions import MissingBackend


class BaseBackendUtilsTest(unittest.TestCase):
    def setUp(self):
        self.strategy = TestStrategy(storage=TestStorage)

    def tearDown(self):
        self.strategy = None


class LoadBackendsTest(BaseBackendUtilsTest):
    def test_load_backends(self):
        loaded_backends = load_backends((
            'social_core.backends.github.GithubOAuth2',
            'social_core.backends.facebook.FacebookOAuth2',
            'social_core.backends.flickr.FlickrOAuth'
        ), force_load=True)
        keys = list(loaded_backends.keys())
        self.assertEqual(keys, ['github', 'facebook', 'flickr'])

        backends = ()
        loaded_backends = load_backends(backends, force_load=True)
        self.assertEqual(len(list(loaded_backends.keys())), 0)


class GetBackendTest(BaseBackendUtilsTest):
    def test_get_backend(self):
        backend = get_backend((
            'social_core.backends.github.GithubOAuth2',
            'social_core.backends.facebook.FacebookOAuth2',
            'social_core.backends.flickr.FlickrOAuth'
        ), 'github')
        self.assertEqual(backend, GithubOAuth2)

    def test_get_missing_backend(self):
        with self.assertRaisesRegex(MissingBackend,
                                    'Missing backend "foobar" entry'):
            get_backend(('social_core.backends.github.GithubOAuth2',
                         'social_core.backends.facebook.FacebookOAuth2',
                         'social_core.backends.flickr.FlickrOAuth'),
                        'foobar')



import json

from .oauth import OAuth2Test


class FlatOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.flat.FlatOAuth2'
    user_data_url = 'https://api.flat.io/v2/me'
    expected_username = 'vincent'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'id': '541a137946fd04d57cb2e3c0',
        'username': 'vincent',
        'name': 'Vincent Foo',
        'printableName': 'Vincent Foo',
        'bio': 'Foo bio',
        'instruments': [],
        'picture': 'https://flat-prod-public.s3.amazonaws.com/00000000/a0d2cb86-ab1e-4fdb-9286-dd94aa6d386c.jpeg',
        'registrationDate': '2014-09-17T23:04:25.042Z',
        'htmlUrl': 'https://flat.io/vincent',
        'starredScoresCount': 85,
        'likedScoresCount': 85,
        'followersCount': 183,
        'followingCount': 52,
        'ownedPublicScoresCount': 15,
        'isPowerUser': True
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class QiitaOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.qiita.QiitaOAuth2'
    user_data_url = 'https://qiita.com/api/v2/authenticated_user'
    expected_username = 'foobar'

    access_token_body = json.dumps({
        'token': 'foobar',
        'token_type': 'bearer'
    })

    user_data_body = json.dumps({
        'id': 'foobar',
        'name': 'Foo Bar'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json
from urllib.parse import urlencode

from httpretty import HTTPretty

from ...exceptions import AuthForbidden
from .oauth import OAuth1Test, OAuth2Test


class BitbucketOAuthMixin:
    user_data_url = 'https://api.bitbucket.org/2.0/user'
    expected_username = 'foobar'
    bb_api_user_emails = 'https://api.bitbucket.org/2.0/user/emails'

    user_data_body = json.dumps({
        u'created_on': u'2012-03-29T18:07:38+00:00',
        u'display_name': u'Foo Bar',
        u'links': {
            u'avatar': {u'href': u'https://bitbucket.org/account/foobar/avatar/32/'},
            u'followers': {u'href': u'https://api.bitbucket.org/2.0/users/foobar/followers'},
            u'following': {u'href': u'https://api.bitbucket.org/2.0/users/foobar/following'},
            u'hooks': {u'href': u'https://api.bitbucket.org/2.0/users/foobar/hooks'},
            u'html': {u'href': u'https://bitbucket.org/foobar'},
            u'repositories': {u'href': u'https://api.bitbucket.org/2.0/repositories/foobar'},
            u'self': {u'href': u'https://api.bitbucket.org/2.0/users/foobar'}},
        u'location': u'Fooville, Bar',
        u'type': u'user',
        u'username': u'foobar',
        u'uuid': u'{397621dc-0f78-329f-8d6d-727396248e3f}',
        u'website': u'http://foobar.com'
    })

    emails_body = json.dumps({
        u'page': 1,
        u'pagelen': 10,
        u'size': 2,
        u'values': [
            {
                u'email': u'foo@bar.com',
                u'is_confirmed': True,
                u'is_primary': True,
                u'links': {u'self': {u'href': u'https://api.bitbucket.org/2.0/user/emails/foo@bar.com'}},
                u'type': u'email'
            },
            {
                u'email': u'not@confirme.com',
                u'is_confirmed': False,
                u'is_primary': False,
                u'links': {u'self': {u'href': u'https://api.bitbucket.org/2.0/user/emails/not@confirmed.com'}},
                u'type': u'email'
            }
        ]
    })


class BitbucketOAuth1Test(BitbucketOAuthMixin, OAuth1Test):
    backend_path = 'social_core.backends.bitbucket.BitbucketOAuth'

    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })

    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })

    def test_login(self):
        HTTPretty.register_uri(HTTPretty.GET,
                               self.bb_api_user_emails,
                               status=200, body=self.emails_body)
        self.do_login()

    def test_partial_pipeline(self):
        HTTPretty.register_uri(HTTPretty.GET,
                               self.bb_api_user_emails,
                               status=200, body=self.emails_body)
        self.do_partial_pipeline()


class BitbucketOAuth1FailTest(BitbucketOAuth1Test):
    emails_body = json.dumps({
        u'page': 1,
        u'pagelen': 10,
        u'size': 1,
        u'values': [
            {
                u'email': u'foo@bar.com',
                u'is_confirmed': False,
                u'is_primary': True,
                u'links': {u'self': {u'href': u'https://api.bitbucket.org/2.0/user/emails/foo@bar.com'}},
                u'type': u'email'
            }
        ]
    })

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_BITBUCKET_VERIFIED_EMAILS_ONLY': True
        })
        with self.assertRaises(AuthForbidden):
            super().test_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_BITBUCKET_VERIFIED_EMAILS_ONLY': True
        })
        with self.assertRaises(AuthForbidden):
            super().test_partial_pipeline()


class BitbucketOAuth2Test(BitbucketOAuthMixin, OAuth2Test):
    backend_path = 'social_core.backends.bitbucket.BitbucketOAuth2'

    access_token_body = json.dumps({
        'access_token': 'foobar_access',
        'scopes': 'foo_scope',
        'expires_in': 3600,
        'refresh_token': 'foobar_refresh',
        'token_type': 'bearer'
    })

    def test_login(self):
        HTTPretty.register_uri(HTTPretty.GET,
                               self.bb_api_user_emails,
                               status=200, body=self.emails_body)
        self.do_login()

    def test_partial_pipeline(self):
        HTTPretty.register_uri(HTTPretty.GET,
                               self.bb_api_user_emails,
                               status=200, body=self.emails_body)
        self.do_partial_pipeline()


class BitbucketOAuth2FailTest(BitbucketOAuth2Test):
    emails_body = json.dumps({
        u'page': 1,
        u'pagelen': 10,
        u'size': 1,
        u'values': [
            {
                u'email': u'foo@bar.com',
                u'is_confirmed': False,
                u'is_primary': True,
                u'links': {u'self': {u'href': u'https://api.bitbucket.org/2.0/user/emails/foo@bar.com'}},
                u'type': u'email'
            }
        ]
    })

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_BITBUCKET_OAUTH2_VERIFIED_EMAILS_ONLY': True
        })
        with self.assertRaises(AuthForbidden):
            super().test_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_BITBUCKET_OAUTH2_VERIFIED_EMAILS_ONLY': True
        })
        with self.assertRaises(AuthForbidden):
            super().test_partial_pipeline()



import json

from ...backends.facebook import API_VERSION
from ...exceptions import AuthUnknownError, AuthCanceled

from .oauth import OAuth2Test


class FacebookOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.facebook.FacebookOAuth2'
    user_data_url = 'https://graph.facebook.com/v{version}/me'.format(
        version=API_VERSION
    )
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'username': 'foobar',
        'first_name': 'Foo',
        'last_name': 'Bar',
        'verified': True,
        'name': 'Foo Bar',
        'gender': 'male',
        'updated_time': '2013-02-13T14:59:42+0000',
        'link': 'http://www.facebook.com/foobar',
        'id': '110011001100010'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class FacebookOAuth2WrongUserDataTest(FacebookOAuth2Test):
    user_data_body = 'null'

    def test_login(self):
        with self.assertRaises(AuthUnknownError):
            self.do_login()

    def test_partial_pipeline(self):
        with self.assertRaises(AuthUnknownError):
            self.do_partial_pipeline()


class FacebookOAuth2AuthCancelTest(FacebookOAuth2Test):
    access_token_status = 400
    access_token_body = json.dumps({
        'error': {
            'message': "redirect_uri isn't an absolute URI. Check RFC 3986.",
            'code': 191,
            'type': 'OAuthException',
            'fbtrace_id': '123Abc'
        }
    })

    def test_login(self):
        with self.assertRaises(AuthCanceled) as cm:
            self.do_login()
        self.assertIn('error', cm.exception.response.json())

    def test_partial_pipeline(self):
        with self.assertRaises(AuthCanceled) as cm:
            self.do_partial_pipeline()
        self.assertIn('error', cm.exception.response.json())



"""
Drip OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/drip.html
"""
from .oauth import BaseOAuth2


class DripOAuth(BaseOAuth2):
    name = 'drip'
    AUTHORIZATION_URL = 'https://www.getdrip.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://www.getdrip.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'

    def get_user_id(self, details, response):
        return details['email']

    def get_user_details(self, response):
        return {'email': response['users'][0]['email'],
                'fullname': response['users'][0]['name'],
                'username': response['users'][0]['email']}

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json('https://api.getdrip.com/v2/user', headers={
            'Authorization': 'Bearer %s' % access_token
        })



"""
OpenStack OpenId backend
"""
from urllib.parse import urlsplit
from openid.extensions import ax

from .open_id import OpenIdAuth


class OpenStackOpenId(OpenIdAuth):
    name = 'openstack'
    URL = 'openstackid.org'

    def get_user_details(self, response):
        """Generate username from identity url"""
        values = super().get_user_details(response)
        values['username'] = values.get('username') or \
            urlsplit(response.identity_url).path.strip('/')
        values['nickname'] = values.get('nickname', '')
        return values

    def setup_request(self, params=None):
        """Fetch email, firstname, lastname from openid"""
        request = self.openid_request(params)

        # TODO: use sreg instead ax request to fetch nickname as username
        fetch_request = ax.FetchRequest()
        fetch_request.add(ax.AttrInfo(
            'http://axschema.org/contact/email',
            alias='email',
            required=True
        ))

        fetch_request.add(ax.AttrInfo(
            'http://axschema.org/namePerson/first',
            alias='firstname',
            required=True
        ))

        fetch_request.add(ax.AttrInfo(
            'http://axschema.org/namePerson/last',
            alias='lastname',
            required=True
        ))

        request.addExtension(fetch_request)
        return request



"""
Okta OAuth2 and OpenIdConnect:
    https://python-social-auth.readthedocs.io/en/latest/backends/okta.html
"""
from urllib.parse import urljoin

from ..utils import append_slash
from .oauth import BaseOAuth2


class OktaMixin:
    def api_url(self):
        return append_slash(self.setting('API_URL'))

    def authorization_url(self):
        return self._url('v1/authorize')

    def access_token_url(self):
        return self._url('v1/token')

    def _url(self, path):
        return urljoin(append_slash(self.setting('API_URL')), path)

    def oidc_config(self):
        return self.get_json(
            self._url(
                './.well-known/openid-configuration?client_id={}'.format(
                    self.setting('KEY')
                )
            )
        )


class OktaOAuth2(OktaMixin, BaseOAuth2):
    """Okta OAuth authentication backend"""
    name = 'okta-oauth2'
    REDIRECT_STATE = False
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ' '
    ID_KEY = 'preferred_username'

    DEFAULT_SCOPE = [
        'openid', 'profile', 'email'
    ]
    EXTRA_DATA = [
        ('refresh_token', 'refresh_token', True),
        ('expires_in', 'expires'),
        ('token_type', 'token_type', True)
    ]

    def get_user_details(self, response):
        """Return user details from Okta account"""
        return {'username': response.get('preferred_username'),
                'email': response.get('email') or '',
                'first_name': response.get('given_name'),
                'last_name': response.get('family_name')}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from Okta"""
        return self.get_json(
            self._url('v1/userinfo'),
            headers={
                'Authorization': 'Bearer {}'.format(access_token),
            }
        )



import json

from .oauth import OAuth2Test


class LyftOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.lyft.LyftOAuth2'
    user_data_url = 'https://api.lyft.com/v1/profile'
    access_token_body = json.dumps({
        'access_token': 'atoken_foo',
        'refresh_token': 'rtoken_bar',
        'token_type': 'bearer',
        'expires_in': 3600,
        'scope': 'public profile rides.read rides.request',
        'id': 'user_foobar'
    })
    user_data_body = json.dumps({
        'id': 'user_foobar'
    })
    expected_username = 'user_foobar'

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Instagram OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/instagram.html
"""
from .oauth import BaseOAuth2


class InstagramOAuth2(BaseOAuth2):
    name = 'instagram'
    AUTHORIZATION_URL = 'https://api.instagram.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://api.instagram.com/oauth/access_token'
    ACCESS_TOKEN_METHOD = 'POST'

    def get_user_id(self, details, response):
        user = response.get('user') or {}
        return user.get('id')

    def get_user_details(self, response):
        """Return user details from Instagram account"""
        user = response.get('user') or {}
        username = user['username']
        email = user.get('email', '')
        fullname, first_name, last_name = self.get_user_names(
            user.get('full_name', '')
        )
        return {'username': username,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': email}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        # For more fields see:
        # https://developers.facebook.com/docs/instagram-basic-display-api/reference/user#fields
        # In fact there are not very many of them.
        fields = 'id,username'
        params = {'access_token': access_token, 'fields': fields}
        response = self.get_json('https://graph.instagram.com/me',
                             params=params)
        return {'user': response}

    def auth_html(self):
        pass



import json
import datetime
from calendar import timegm

from jose import jwk, jwt
from jose.jwt import JWTError, JWTClaimsError, ExpiredSignatureError
from jose.utils import base64url_decode

from social_core.backends.oauth import BaseOAuth2
from social_core.utils import cache
from social_core.exceptions import AuthTokenError


class OpenIdConnectAssociation:
    """ Use Association model to save the nonce by force."""

    def __init__(self, handle, secret='', issued=0, lifetime=0, assoc_type=''):
        self.handle = handle  # as nonce
        self.secret = secret.encode()  # not use
        self.issued = issued  # not use
        self.lifetime = lifetime  # not use
        self.assoc_type = assoc_type  # as state


class OpenIdConnectAuth(BaseOAuth2):
    """
    Base class for Open ID Connect backends.
    Currently only the code response type is supported.
    """
    # Override OIDC_ENDPOINT in your subclass to enable autoconfig of OIDC
    OIDC_ENDPOINT = None
    ID_TOKEN_MAX_AGE = 600
    DEFAULT_SCOPE = ['openid', 'profile', 'email']
    EXTRA_DATA = ['id_token', 'refresh_token', ('sub', 'id')]
    REDIRECT_STATE = False
    ACCESS_TOKEN_METHOD = 'POST'
    REVOKE_TOKEN_METHOD = 'GET'
    ID_KEY = 'sub'
    USERNAME_KEY = 'preferred_username'
    ID_TOKEN_ISSUER = ''
    ACCESS_TOKEN_URL = ''
    AUTHORIZATION_URL = ''
    REVOKE_TOKEN_URL = ''
    USERINFO_URL = ''
    JWKS_URI = ''
    JWT_ALGORITHMS = ['RS256']
    JWT_DECODE_OPTIONS = dict()

    def __init__(self, *args, **kwargs):
        self.id_token = None
        super().__init__(*args, **kwargs)

    def authorization_url(self):
        return self.AUTHORIZATION_URL or \
            self.oidc_config().get('authorization_endpoint')

    def access_token_url(self):
        return self.ACCESS_TOKEN_URL or \
            self.oidc_config().get('token_endpoint')

    def revoke_token_url(self, token, uid):
        return self.REVOKE_TOKEN_URL or \
            self.oidc_config().get('revocation_endpoint')

    def id_token_issuer(self):
        return self.ID_TOKEN_ISSUER or \
            self.oidc_config().get('issuer')

    def userinfo_url(self):
        return self.USERINFO_URL or \
            self.oidc_config().get('userinfo_endpoint')

    def jwks_uri(self):
        return self.JWKS_URI or \
            self.oidc_config().get('jwks_uri')

    @cache(ttl=86400)
    def oidc_config(self):
        return self.get_json(self.OIDC_ENDPOINT +
                             '/.well-known/openid-configuration')

    @cache(ttl=86400)
    def get_jwks_keys(self):
        keys = self.get_remote_jwks_keys()

        # Add client secret as oct key so it can be used for HMAC signatures
        # client_id, client_secret = self.get_key_and_secret()
        # keys.append({'key': client_secret, 'kty': 'oct'})
        return keys

    def get_remote_jwks_keys(self):
        response = self.request(self.jwks_uri())
        return json.loads(response.text)['keys']

    def auth_params(self, state=None):
        """Return extra arguments needed on auth process."""
        params = super().auth_params(state)
        params['nonce'] = self.get_and_store_nonce(
            self.authorization_url(), state
        )
        return params

    def get_and_store_nonce(self, url, state):
        # Create a nonce
        nonce = self.strategy.random_string(64)
        # Store the nonce
        association = OpenIdConnectAssociation(nonce, assoc_type=state)
        self.strategy.storage.association.store(url, association)
        return nonce

    def get_nonce(self, nonce):
        try:
            return self.strategy.storage.association.get(
                server_url=self.authorization_url(),
                handle=nonce
            )[0]
        except IndexError:
            pass

    def remove_nonce(self, nonce_id):
        self.strategy.storage.association.remove([nonce_id])

    def validate_claims(self, id_token):
        utc_timestamp = timegm(datetime.datetime.utcnow().utctimetuple())

        if 'nbf' in id_token and utc_timestamp < id_token['nbf']:
            raise AuthTokenError(self, 'Incorrect id_token: nbf')

        # Verify the token was issued in the last 10 minutes
        iat_leeway = self.setting('ID_TOKEN_MAX_AGE', self.ID_TOKEN_MAX_AGE)
        if utc_timestamp > id_token['iat'] + iat_leeway:
            raise AuthTokenError(self, 'Incorrect id_token: iat')

        # Validate the nonce to ensure the request was not modified
        nonce = id_token.get('nonce')
        if not nonce:
            raise AuthTokenError(self, 'Incorrect id_token: nonce')

        nonce_obj = self.get_nonce(nonce)
        if nonce_obj:
            self.remove_nonce(nonce_obj.id)
        else:
            raise AuthTokenError(self, 'Incorrect id_token: nonce')

    def find_valid_key(self, id_token):
        for key in self.get_jwks_keys():
            rsakey = jwk.construct(key)
            message, encoded_sig = id_token.rsplit('.', 1)
            decoded_sig = base64url_decode(encoded_sig.encode('utf-8'))
            if rsakey.verify(message.encode('utf-8'), decoded_sig):
                return key

    def validate_and_return_id_token(self, id_token, access_token):
        """
        Validates the id_token according to the steps at
        http://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation.
        """
        client_id, client_secret = self.get_key_and_secret()

        key = self.find_valid_key(id_token)

        if not key:
            raise AuthTokenError(self, 'Signature verification failed')

        rsakey = jwk.construct(key)

        try:
            claims = jwt.decode(
                id_token,
                rsakey.to_pem().decode('utf-8'),
                algorithms=self.JWT_ALGORITHMS,
                audience=client_id,
                issuer=self.id_token_issuer(),
                access_token=access_token,
                options=self.JWT_DECODE_OPTIONS,
            )
        except ExpiredSignatureError:
            raise AuthTokenError(self, 'Signature has expired')
        except JWTClaimsError as error:
            raise AuthTokenError(self, str(error))
        except JWTError:
            raise AuthTokenError(self, 'Invalid signature')

        self.validate_claims(claims)

        return claims

    def request_access_token(self, *args, **kwargs):
        """
        Retrieve the access token. Also, validate the id_token and
        store it (temporarily).
        """
        response = self.get_json(*args, **kwargs)
        self.id_token = self.validate_and_return_id_token(
            response['id_token'],
            response['access_token']
        )
        return response

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(self.userinfo_url(), headers={
            'Authorization': 'Bearer {0}'.format(access_token)
        })

    def get_user_details(self, response):
        username_key = self.setting('USERNAME_KEY', default=self.USERNAME_KEY)
        return {
            'username': response.get(username_key),
            'email': response.get('email'),
            'fullname': response.get('name'),
            'first_name': response.get('given_name'),
            'last_name': response.get('family_name'),
        }



"""
Dribbble OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/dribbble.html
    http://developer.dribbble.com/v1/oauth/
"""

from .oauth import BaseOAuth2


class DribbbleOAuth2(BaseOAuth2):
    """Dribbble OAuth authentication backend"""
    name = 'dribbble'
    AUTHORIZATION_URL = 'https://dribbble.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://dribbble.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ','
    EXTRA_DATA = [
        ('id', 'id'),
        ('name', 'name'),
        ('html_url', 'html_url'),
        ('avatar_url', 'avatar_url'),
        ('bio', 'bio'),
        ('location', 'location'),
        ('links', 'links'),
        ('buckets_count', 'buckets_count'),
        ('comments_received_count', 'comments_received_count'),
        ('followers_count', 'followers_count'),
        ('followings_count', 'followings_count'),
        ('likes_count', 'likes_count'),
        ('likes_received_count', 'likes_received_count'),
        ('projects_count', 'projects_count'),
        ('rebounds_received_count', 'rebounds_received_count'),
        ('shots_count', 'shots_count'),
        ('teams_count', 'teams_count'),
        ('pro', 'pro'),
        ('buckets_url', 'buckets_url'),
        ('followers_url', 'followers_url'),
        ('following_url', 'following_url'),
        ('likes_url', 'shots_url'),
        ('teams_url', 'teams_url'),
        ('created_at', 'created_at'),
        ('updated_at', 'updated_at'),
    ]

    def get_user_details(self, response):
        """Return user details from Dribbble account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('name')
        )
        return {'username': response.get('username'),
                'email': response.get('email', ''),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://api.dribbble.com/v1/user',
            headers={
                'Authorization': 'Bearer {0}'.format(access_token)
            })



import json

from .oauth import OAuth2Test


class BaseLinkedinTest:
    user_data_url = 'https://api.linkedin.com/v2/me' \
                    '?projection=(firstName,id,lastName)'
    expected_username = 'FooBar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })

    # Reference:
    # https://docs.microsoft.com/en-us/linkedin/consumer/integrations/self
    # -serve/sign-in-with-linkedin?context=linkedin/consumer/context#api-request
    user_data_body = json.dumps({
        'id': '1010101010',
        'firstName': {
            'localized': {
                'en_US': 'Foo'
            },
            'preferredLocale': {
                'country': 'US',
                'language': 'en'
            }
        },
        'lastName': {
            'localized': {
                'en_US': 'Bar'
            },
            'preferredLocale': {
                'country': 'US',
                'language': 'en'
            }
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class LinkedinOAuth2Test(BaseLinkedinTest, OAuth2Test):
    backend_path = 'social_core.backends.linkedin.LinkedinOAuth2'


class LinkedinMobileOAuth2Test(BaseLinkedinTest, OAuth2Test):
    backend_path = 'social_core.backends.linkedin.LinkedinMobileOAuth2'



from urllib.parse import urlencode

from requests import HTTPError

from ...exceptions import AuthCanceled
from .oauth import OAuth1Test


class EvernoteOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.evernote.EvernoteOAuth'
    expected_username = '101010'
    access_token_body = urlencode({
        'edam_webApiUrlPrefix': 'https://sandbox.evernote.com/shard/s1/',
        'edam_shard': 's1',
        'oauth_token': 'foobar',
        'edam_expires': '1395118279645',
        'edam_userId': '101010',
        'edam_noteStoreUrl': 'https://sandbox.evernote.com/shard/s1/notestore'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class EvernoteOAuth1CanceledTest(EvernoteOAuth1Test):
    access_token_status = 401

    def test_login(self):
        with self.assertRaises(AuthCanceled) as cm:
            self.do_login()
        self.assertTrue(cm.exception.response is not None)

    def test_partial_pipeline(self):
        with self.assertRaises(AuthCanceled) as cm:
            self.do_partial_pipeline()
        self.assertTrue(cm.exception.response is not None)


class EvernoteOAuth1ErrorTest(EvernoteOAuth1Test):
    access_token_status = 500

    def test_login(self):
        with self.assertRaises(HTTPError):
            self.do_login()

    def test_partial_pipeline(self):
        with self.assertRaises(HTTPError):
            self.do_partial_pipeline()



from .oauth import BaseOAuth2


class WunderlistOAuth2(BaseOAuth2):
    """Wunderlist OAuth2 authentication backend"""
    name = 'wunderlist'
    AUTHORIZATION_URL = 'https://www.wunderlist.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://www.wunderlist.com/oauth/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def get_user_details(self, response):
        """Return user details from Wunderlist account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('name')
        )
        return {'username': str(response.get('id')),
                'email': response.get('email'),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        headers = {
            'X-Access-Token': access_token,
            'X-Client-ID': self.setting('KEY')}
        return self.get_json(
            'https://a.wunderlist.com/api/v1/user', headers=headers)



import json

from .oauth import OAuth2Test


class BoxOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.box.BoxOAuth2'
    user_data_url = 'https://api.box.com/2.0/users/me'
    expected_username = 'sean+awesome@box.com'
    access_token_body = json.dumps({
        'access_token': 'T9cE5asGnuyYCCqIZFoWjFHvNbvVqHjl',
        'expires_in': 3600,
        'restricted_to': [],
        'token_type': 'bearer',
        'refresh_token': 'J7rxTiWOHMoSC1isKZKBZWizoRXjkQzig5C6jFgCVJ9bU'
                         'nsUfGMinKBDLZWP9BgR'
    })
    user_data_body = json.dumps({
        'type': 'user',
        'id': '181216415',
        'name': 'sean rose',
        'login': 'sean+awesome@box.com',
        'created_at': '2012-05-03T21:39:11-07:00',
        'modified_at': '2012-11-14T11:21:32-08:00',
        'role': 'admin',
        'language': 'en',
        'space_amount': 11345156112,
        'space_used': 1237009912,
        'max_upload_size': 2147483648,
        'tracking_codes': [],
        'can_see_managed_users': True,
        'is_sync_enabled': True,
        'status': 'active',
        'job_title': '',
        'phone': '6509241374',
        'address': '',
        'avatar_url': 'https://www.box.com/api/avatar/large/181216415',
        'is_exempt_from_device_limits': False,
        'is_exempt_from_login_verification': False,
        'enterprise': {
            'type': 'enterprise',
            'id': '17077211',
            'name': 'seanrose enterprise'
        }
    })
    refresh_token_body = json.dumps({
        'access_token': 'T9cE5asGnuyYCCqIZFoWjFHvNbvVqHjl',
        'expires_in': 3600,
        'restricted_to': [],
        'token_type': 'bearer',
        'refresh_token': 'J7rxTiWOHMoSC1isKZKBZWizoRXjkQzig5C6jFgCVJ9b'
                         'UnsUfGMinKBDLZWP9BgR'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def refresh_token_arguments(self):
        uri = self.strategy.build_absolute_uri('/complete/box/')
        return {'redirect_uri': uri}

    def test_refresh_token(self):
        user, social = self.do_refresh_token()
        self.assertEqual(social.extra_data['access_token'],
                         'T9cE5asGnuyYCCqIZFoWjFHvNbvVqHjl')



import json

from .oauth import OAuth2Test


class CoinbaseOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.coinbase.CoinbaseOAuth2'
    user_data_url = 'https://api.coinbase.com/v2/user'
    expected_username = 'satoshi_nakomoto'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'data': {
            'id': '1A1zP1eP5QGefi2DMPTfTL5SLmv7DivfNa',
            'name': 'Satoshi Nakamoto',
            'username': 'satoshi_nakomoto',
            'profile_location': None,
            'profile_bio': None,
            'profile_url': 'https://coinbase.com/satoshi_nakomoto',
            'avatar_url': None,
            'resource': 'user',
            'resource_path': '/v2/user'
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class BehanceOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.behance.BehanceOAuth2'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'valid': 1,
        'user': {
            'username': 'foobar',
            'city': 'Foo City',
            'first_name': 'Foo',
            'last_name': 'Bar',
            'display_name': 'Foo Bar',
            'url': 'http://www.behance.net/foobar',
            'country': 'Fooland',
            'company': '',
            'created_on': 1355152329,
            'state': '',
            'fields': [
                'Programming',
                'Web Design',
                'Web Development'
            ],
            'images': {
                '32': 'https://www.behance.net/assets/img/profile/'
                      'no-image-32.jpg',
                '50': 'https://www.behance.net/assets/img/profile/'
                      'no-image-50.jpg',
                '115': 'https://www.behance.net/assets/img/profile/'
                       'no-image-138.jpg',
                '129': 'https://www.behance.net/assets/img/profile/'
                       'no-image-138.jpg',
                '138': 'https://www.behance.net/assets/img/profile/'
                       'no-image-138.jpg',
                '78': 'https://www.behance.net/assets/img/profile/'
                      'no-image-78.jpg'
            },
            'id': 1010101,
            'occupation': 'Software Developer'
        }
    })
    expected_username = 'foobar'

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
MapMyFitness OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/mapmyfitness.html
"""
from .oauth import BaseOAuth2


class MapMyFitnessOAuth2(BaseOAuth2):
    """MapMyFitness OAuth authentication backend"""
    name = 'mapmyfitness'
    AUTHORIZATION_URL = 'https://www.mapmyfitness.com/v7.0/oauth2/authorize'
    ACCESS_TOKEN_URL = \
        'https://oauth2-api.mapmyapi.com/v7.0/oauth2/access_token'
    REQUEST_TOKEN_METHOD = 'POST'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('refresh_token', 'refresh_token'),
    ]

    def auth_headers(self):
        key = self.get_key_and_secret()[0]
        return {
            'Api-Key': key
        }

    def get_user_id(self, details, response):
        return response['id']

    def get_user_details(self, response):
        first = response.get('first_name', '')
        last = response.get('last_name', '')
        full = (first + last).strip()
        return {
            'username': response['username'],
            'email': response['email'],
            'fullname': full,
            'first_name': first,
            'last_name': last,
        }

    def user_data(self, access_token, *args, **kwargs):
        key = self.get_key_and_secret()[0]
        url = 'https://oauth2-api.mapmyapi.com/v7.0/user/self/'
        headers = {
            'Authorization': 'Bearer {0}'.format(access_token),
            'Api-Key': key
        }
        return self.get_json(url, headers=headers)



"""
Bungie OAuth2 backend
"""
from social_core.backends.oauth import BaseOAuth2


class BungieOAuth2(BaseOAuth2):
    name = 'bungie'
    ID_KEY = 'membership_id'
    AUTHORIZATION_URL = 'https://www.bungie.net/en/oauth/authorize/'
    ACCESS_TOKEN_URL = 'https://www.bungie.net/platform/app/oauth/token/'
    REFRESH_TOKEN_URL = 'https://www.bungie.net/platform/app/oauth/token/'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('refresh_token', 'refresh_token', True),
        ('access_token', 'access_token', True),
        ('expires_in', 'expires'),
        ('membership_id', 'membership_id'),
        ('refresh_expires_in', 'refresh_expires_in')
    ]

    def auth_html(self):
        """Abstract Method Inclusion"""
        pass

    def auth_headers(self):
        """Adds X-API-KEY and Origin"""
        return {
            'X-API-KEY': self.setting('API_KEY'),
            'Content-Type': 'application/x-www-form-urlencoded',
            'Origin': self.setting('ORIGIN'),
            'Accept': 'application/json'
        }

    def make_bungie_request(self, url, access_token, kwargs):
        """Helper function to get username data keyed off displayName"""
        headers = self.auth_headers()
        headers['Authorization'] = 'Bearer ' + access_token
        return self.get_json(url, headers=headers)

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        state = self.validate_state()
        response = self.request_access_token(
            self.access_token_url(),
            data=self.auth_complete_params(state),
            headers=self.auth_headers(),
            auth=self.auth_complete_credentials(),
            method=self.ACCESS_TOKEN_METHOD
        )
        self.process_error(response)
        return self.do_auth(response['access_token'],
                            response=response,
                            *args, **kwargs)

    def do_auth(self, access_token, *args, **kwargs):
        """Finish the auth process once the access_token was retrieved"""
        data = self.user_data(access_token, *args, **kwargs)
        response = kwargs.get('response') or {}
        response.update(data or {})
        if 'access_token' not in response:
            response['Response']['access_token']['value'] = access_token
        kwargs.update({'response': response, 'backend': self})
        return self.strategy.authenticate(*args, **kwargs)

    def user_data(self, access_token, *args, **kwargs):
        """Grab user profile information from Bunige"""
        membership_id = kwargs['response']['membership_id']
        url = 'https://www.bungie.net/Platform/User/GetBungieNetUser/'
        response = self.make_bungie_request(url, access_token, kwargs)
        username = response['Response']['user']['displayName']
        return {'username': username,
                'uid': membership_id}

    def get_user_details(self, response, *args, **kwargs):
        """Return user details from Bungie account"""
        username = response['username']
        return {
            'first_name': username,
            'username': username,
            'uid': response['uid'],
        }



import hmac
import time
import hashlib

from .base import BaseAuth
from ..exceptions import AuthFailed, AuthMissingParameter
from ..utils import handle_http_errors


class TelegramAuth(BaseAuth):
    name = 'telegram'
    ID_KEY = 'id'

    def verify_data(self, response):
        bot_token = self.setting('BOT_TOKEN')
        if bot_token is None:
            raise AuthMissingParameter('telegram',
                                       'SOCIAL_AUTH_TELEGRAM_BOT_TOKEN')

        received_hash_string = response.get('hash')
        auth_date = response.get('auth_date')

        if received_hash_string is None or auth_date is None:
            raise AuthMissingParameter('telegram', 'hash or auth_date')

        data_check_string = ['{}={}'.format(k, v)
                             for k, v in response.items() if k != 'hash']
        data_check_string = '\n'.join(sorted(data_check_string))
        secret_key = hashlib.sha256(bot_token.encode()).digest()
        built_hash = hmac.new(secret_key,
                              msg=data_check_string.encode(),
                              digestmod=hashlib.sha256).hexdigest()
        current_timestamp = int(time.time())
        auth_timestamp = int(auth_date)
        if current_timestamp - auth_timestamp > 86400:
            raise AuthFailed('telegram', 'Auth date is outdated')
        if built_hash != received_hash_string:
            raise AuthFailed('telegram', 'Invalid hash supplied')

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        return response

    def get_user_details(self, response):
        first_name = response.get('first_name', '')
        last_name = response.get('last_name', '')
        fullname = '{} {}'.format(first_name, last_name).strip()
        return {
            'username': response.get('username') or response[self.ID_KEY],
            'first_name': first_name,
            'last_name': last_name,
            'fullname': fullname
        }

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        response = self.data
        self.verify_data(response)
        kwargs.update({'response': self.data, 'backend': self})
        return self.strategy.authenticate(*args, **kwargs)



import hmac
import time

from base64 import urlsafe_b64decode, urlsafe_b64encode
from hashlib import sha256
from urllib.parse import urlencode

from .base import BaseAuth
from ..exceptions import AuthException, AuthTokenError
from ..utils import parse_qs


class DiscourseAuth(BaseAuth):
    name = 'discourse'
    EXTRA_DATA = ['username', 'name', 'avatar_url']

    def auth_url(self):
        """
        Get the URL to which we must redirect in order to authenticate the user
        """
        return_url = self.redirect_uri
        nonce = self.strategy.random_string(64)
        self.add_nonce(nonce)

        payload = urlencode({
            'nonce': nonce,
            'return_sso_url': return_url
        })
        base_64_payload = urlsafe_b64encode(
            payload.encode('utf8')
        ).decode('ascii')

        payload_signature = hmac.new(
            self.setting('SECRET').encode('utf8'),
            base_64_payload.encode('utf8'),
            sha256,
        ).hexdigest()
        encoded_params = urlencode({
            'sso': base_64_payload,
            'sig': payload_signature
        })
        return '{0}?{1}'.format(self.get_idp_url(), encoded_params)

    def get_idp_url(self):
        return self.setting('SERVER_URL') + '/session/sso_provider'

    def get_user_id(self, details, response):
        return response['email']

    def get_user_details(self, response):
        results = {
            'username': response.get('username'),
            'email': response.get('email'),
            'name': response.get('name'),
            'groups': response.get('groups', '').split(','),
            'is_staff': response.get('admin') == 'true' or
                        response.get('moderator') == 'true',
            'is_superuser': response.get('admin') == 'true',
        }
        return results

    def add_nonce(self, nonce):
        self.strategy.storage.nonce.use(
            self.setting('SERVER_URL'),
            time.time(),
            nonce
        )

    def get_nonce(self, nonce):
        return self.strategy.storage.nonce.get(
            self.setting('SERVER_URL'),
            nonce
        )

    def delete_nonce(self, nonce):
        self.strategy.storage.nonce.delete(nonce)

    def auth_complete(self, *args, **kwargs):
        """
        The user has been redirected back from the IdP and we should
        now log them in, if everything checks out.
        """
        request_data = self.strategy.request_data()

        sso_params = request_data.get('sso')
        sso_signature = request_data.get('sig')

        param_signature = hmac.new(
            self.setting('SECRET').encode('utf8'),
            sso_params.encode('utf8'),
            sha256
        ).hexdigest()

        if not hmac.compare_digest(str(sso_signature), str(param_signature)):
            raise AuthException('Could not verify discourse login')

        decoded_params = urlsafe_b64decode(
            sso_params.encode('utf8')
        ).decode('ascii')

        # Validate the nonce to ensure the request was not modified
        response = parse_qs(decoded_params)
        nonce_obj = self.get_nonce(response.get('nonce'))
        if nonce_obj:
            self.delete_nonce(nonce_obj)
        else:
            raise AuthTokenError(self, 'Incorrect id_token: nonce')

        kwargs.update({
            'sso': '',
            'sig': '',
            'backend': self,
            'response': response
        })
        return self.strategy.authenticate(*args, **kwargs)



import json

from .oauth import OAuth2Test
from social_core.backends.paypal import PayPalOAuth2


class PayPalOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.paypal.PayPalOAuth2'
    user_data_url = (
        'https://api.paypal.com/v1/identity/oauth2/userinfo?schema=paypalv1.1'
    )
    expected_username = 'mWq6_1sU85v5EG9yHdPxJRrhGHrnMJ-1PQKtX6pcsmA'
    access_token_body = json.dumps(
        {
            'token_type': 'Bearer',
            'expires_in': 28800,
            'refresh_token': 'foobar-refresh-token',
            'access_token': 'foobar-token',
        }
    )
    user_data_body = json.dumps(
        {
            'user_id': 'https://www.paypal.com/webapps/auth/identity/user/mWq6_1sU85v5EG9yHdPxJRrhGHrnMJ-1PQKtX6pcsmA',
            'name': 'identity test',
            'given_name': 'identity',
            'family_name': 'test',
            'payer_id': 'WDJJHEBZ4X2LY',
            'address': {
                'street_address': '1 Main St',
                'locality': 'San Jose',
                'region': 'CA',
                'postal_code': '95131',
                'country': 'US',
            },
            'verified_account': True,
            'emails': [{'value': 'user1@example.com', 'primary': True}],
        }
    )
    refresh_token_body = json.dumps(
        {
            'access_token': 'foobar-new-token',
            'token_type': 'Bearer',
            'refresh_token': 'foobar-new-refresh-token',
            'expires_in': 28800,
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_refresh_token(self):
        user, social = self.do_refresh_token()
        self.assertEqual(user.username, self.expected_username)
        self.assertEqual(social.extra_data['access_token'], 'foobar-new-token')

    def test_get_email_no_emails(self):
        emails = []
        email = PayPalOAuth2.get_email(emails)
        self.assertEqual(email, '')

    def test_get_email_multiple_emails(self):
        expected_email = 'mail2@example.com'
        emails = [
            {'value': 'mail1@example.com', 'primary': False},
            {'value': expected_email, 'primary': True},
        ]
        email = PayPalOAuth2.get_email(emails)
        self.assertEqual(email, expected_email)

    def test_get_email_multiple_emails_no_primary(self):
        expected_email = 'mail1@example.com'
        emails = [
            {'value': expected_email, 'primary': False},
            {'value': 'mail2@example.com', 'primary': False},
        ]
        email = PayPalOAuth2.get_email(emails)
        self.assertEqual(email, expected_email)



import json
from urllib.parse import urlencode

from .oauth import OAuth1Test


class UpworkOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.upwork.UpworkOAuth'
    user_data_url = 'https://www.upwork.com/api/auth/v1/info.json'
    expected_username = '10101010'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    user_data_body = json.dumps({
        'info': {
            'portrait_32_img': '',
            'capacity': {
                'buyer': 'no',
                'affiliate_manager': 'no',
                'provider': 'yes'
            },
            'company_url': '',
            'has_agency': '1',
            'portrait_50_img': '',
            'portrait_100_img': '',
            'location': {
                'city': 'New York',
                'state': '',
                'country': 'USA'
            },
            'ref': '9755314',
            'profile_url': 'https://www.upwork.com/users/~10101010'
        },
        'auth_user': {
            'timezone': 'USA/New York',
            'first_name': 'Foo',
            'last_name': 'Bar',
            'timezone_offset': '10000'
        },
        'server_time': '1111111111'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from urllib.parse import urlparse

import requests
from httpretty import HTTPretty

from ...utils import parse_qs, url_add_parameters
from ..models import User
from .base import BaseBackendTest


class BaseOAuthTest(BaseBackendTest):
    backend = None
    backend_path = None
    user_data_body = None
    user_data_url = ''
    user_data_url_post = False
    user_data_content_type = 'application/json'
    access_token_body = None
    access_token_status = 200
    expected_username = ''

    def extra_settings(self):
        return {'SOCIAL_AUTH_' + self.name + '_KEY': 'a-key',
                'SOCIAL_AUTH_' + self.name + '_SECRET': 'a-secret-key'}

    def _method(self, method):
        return {'GET': HTTPretty.GET,
                'POST': HTTPretty.POST}[method]

    def handle_state(self, start_url, target_url):
        start_query = parse_qs(urlparse(start_url).query)
        redirect_uri = start_query.get('redirect_uri')

        if getattr(self.backend, 'STATE_PARAMETER', False):
            if start_query.get('state'):
                target_url = url_add_parameters(target_url, {
                    'state': start_query['state']
                })

        if redirect_uri and getattr(self.backend, 'REDIRECT_STATE', False):
            redirect_query = parse_qs(urlparse(redirect_uri).query)
            if redirect_query.get('redirect_state'):
                target_url = url_add_parameters(target_url, {
                    'redirect_state': redirect_query['redirect_state']
                })
        return target_url

    def auth_handlers(self, start_url):
        target_url = self.handle_state(start_url,
                                       self.strategy.build_absolute_uri(
                                           self.complete_url
                                       ))
        HTTPretty.register_uri(HTTPretty.GET,
                               start_url,
                               status=301,
                               location=target_url)
        HTTPretty.register_uri(HTTPretty.GET,
                               target_url,
                               status=200,
                               body='foobar')
        HTTPretty.register_uri(self._method(self.backend.ACCESS_TOKEN_METHOD),
                               uri=self.backend.access_token_url(),
                               status=self.access_token_status,
                               body=self.access_token_body or '',
                               content_type='text/json')
        if self.user_data_url:
            HTTPretty.register_uri(HTTPretty.POST if self.user_data_url_post else HTTPretty.GET,
                                   self.user_data_url,
                                   body=self.user_data_body or '',
                                   content_type=self.user_data_content_type)
        return target_url

    def do_start(self):
        start_url = self.backend.start().url
        target_url = self.auth_handlers(start_url)
        response = requests.get(start_url)
        self.assertEqual(response.url, target_url)
        self.assertEqual(response.text, 'foobar')
        self.strategy.set_request_data(parse_qs(urlparse(start_url).query),
                                       self.backend)
        self.strategy.set_request_data(parse_qs(urlparse(target_url).query),
                                       self.backend)
        return self.backend.complete()


class OAuth1Test(BaseOAuthTest):
    request_token_body = None
    raw_complete_url = '/complete/{0}/?oauth_verifier=bazqux&' \
        'oauth_token=foobar'

    def request_token_handler(self):
        HTTPretty.register_uri(self._method(self.backend.REQUEST_TOKEN_METHOD),
                               self.backend.REQUEST_TOKEN_URL,
                               body=self.request_token_body,
                               status=200)

    def do_start(self):
        self.request_token_handler()
        return super().do_start()


class OAuth2Test(BaseOAuthTest):
    raw_complete_url = '/complete/{0}/?code=foobar'
    refresh_token_body = ''

    def refresh_token_arguments(self):
        return {}

    def do_refresh_token(self):
        self.do_login()
        HTTPretty.register_uri(self._method(self.backend.REFRESH_TOKEN_METHOD),
                               self.backend.refresh_token_url(),
                               status=200,
                               body=self.refresh_token_body)
        user = list(User.cache.values())[0]
        social = user.social[0]
        social.refresh_token(strategy=self.strategy,
                             **self.refresh_token_arguments())
        return user, social



# -*- coding: utf-8 -*-
from urllib.parse import urlparse, parse_qs

import requests
from httpretty import HTTPretty

from .base import BaseBackendTest
from ...exceptions import AuthException


class DiscourseTest(BaseBackendTest):
    backend_path = 'social_core.backends.discourse.DiscourseAuth'
    expected_username = 'beepboop'
    raw_complete_url = '/complete/{0}/'

    def post_start(self):
        pass

    def do_start(self):
        self.post_start()
        start = self.backend.start()
        start_url = start.url
        return_url = self.backend.redirect_uri
        # NOTE: This is how we generated sso:
        # sso = b64encode(urlencode({
        #     'email': 'user@example.com',
        #     'username': 'beepboop',
        #     'nonce': '6YRje7xlXhpyeJ6qtvBeTUjHkXo1UCTQmCrzN8GXfja3AoAFk2' + \
        #              'CieDRYgSqMYi4W',
        #     'return_sso_url': 'http://myapp.com'
        # }))
        sso = 'dXNlcm5hbWU9YmVlcGJvb3Ambm9uY2U9NllSamU3eGxYaHB5ZUo2cXR2QmV' + \
              'UVWpIa1hvMVVDVFFtQ3J6TjhHWGZqYTNBb0FGazJDaWVEUllnU3FNWWk0Vy' + \
              'ZlbWFpbD11c2VyJTQwZXhhbXBsZS5jb20mcmV0dXJuX3Nzb191cmw9aHR0c' + \
              'CUzQSUyRiUyRm15YXBwLmNvbQ=='
        # NOTE: the signature was verified using the 'foo' key, like so:
        # hmac.new('foo', sso, sha256).hexdigest()
        sig = '04063f17c99a97b1a765c1e0d7bbb61afb8471d79a39ddcd6af5ba3c93eb10e1'
        response_query_params = 'sso={0}&sig={1}'.format(sso, sig)

        response_url = '{0}?{1}'.format(return_url, response_query_params)
        HTTPretty.register_uri(
            HTTPretty.GET, start_url, status=301, location=response_url
        )
        HTTPretty.register_uri(
            HTTPretty.GET,
            return_url,
            status=200,
            content_type='text/html',
        )

        response = requests.get(start_url)
        query_values = dict(
            (k, v[0]) for k, v in parse_qs(urlparse(response.url).query).items()
        )
        self.strategy.set_request_data(query_values, self.backend)

        return self.backend.complete()

    def test_login(self):
        """
        Test that we can authenticate with the Discourse IdP
        """
        # pretend we've started with a URL like /login/discourse:
        self.strategy.set_settings(
            {'SERVER_URL': 'http://example.com', 'SECRET': 'foo'}
        )
        self.do_login()

    def test_failed_login(self):
        """
        Test that authentication fails when our request is signed with a
        different secret than our payload
        """
        self.strategy.set_settings(
            {'SERVER_URL': 'http://example.com', 'SECRET': 'bar'}
        )
        with self.assertRaises(AuthException):
            self.do_login()



import json

from .oauth import OAuth2Test


class NationBuilderOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.nationbuilder.NationBuilderOAuth2'
    user_data_url = 'https://foobar.nationbuilder.com/api/v1/people/me'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
        'created_at': 1422937981,
        'expires_in': 2592000
    })
    user_data_body = json.dumps({
        'person': {
            'twitter_followers_count': None,
            'last_name': 'Bar',
            'rule_violations_count': 0,
            'linkedin_id': None,
            'recruiter_id': None,
            'membership_expires_at': None,
            'donations_raised_count': 0,
            'last_contacted_at': None,
            'prefix': None,
            'profile_content_html': None,
            'email4': None,
            'email2': None,
            'availability': None,
            'occupation': None,
            'user_submitted_address': None,
            'could_vote_status': None,
            'state_upper_district': None,
            'salesforce_id': None,
            'van_id': None,
            'phone_time': None,
            'profile_content': None,
            'auto_import_id': None,
            'parent_id': None,
            'email4_is_bad': False,
            'twitter_updated_at': None,
            'email3_is_bad': False,
            'bio': None,
            'party_member': None,
            'unsubscribed_at': None,
            'fax_number': None,
            'last_contacted_by': None,
            'active_customer_expires_at': None,
            'federal_donotcall': False,
            'warnings_count': 0,
            'first_supporter_at': '2015-02-02T19:30:28-08:00',
            'previous_party': None,
            'donations_raised_amount_this_cycle_in_cents': 0,
            'call_status_name': None,
            'marital_status': None,
            'facebook_updated_at': None,
            'donations_count': 0,
            'note_updated_at': None,
            'closed_invoices_count': None,
            'profile_headline': None,
            'fire_district': None,
            'mobile_normalized': None,
            'import_id': None,
            'last_call_id': None,
            'donations_raised_amount_in_cents': 0,
            'facebook_address': None,
            'is_profile_private': False,
            'last_rule_violation_at': None,
            'sex': None,
            'full_name': 'Foo Bar',
            'last_donated_at': None,
            'donations_pledged_amount_in_cents': 0,
            'primary_email_id': 1,
            'media_market_name': None,
            'capital_amount_in_cents': 500,
            'datatrust_id': None,
            'precinct_code': None,
            'email3': None,
            'religion': None,
            'first_prospect_at': None,
            'judicial_district': None,
            'donations_count_this_cycle': 0,
            'work_address': None,
            'is_twitter_follower': False,
            'email1': 'foobar@gmail.com',
            'email': 'foobar@gmail.com',
            'contact_status_name': None,
            'mobile_opt_in': True,
            'twitter_description': None,
            'parent': None,
            'tags': [],
            'first_volunteer_at': None,
            'inferred_support_level': None,
            'banned_at': None,
            'first_invoice_at': None,
            'donations_raised_count_this_cycle': 0,
            'is_donor': False,
            'twitter_location': None,
            'email1_is_bad': False,
            'legal_name': None,
            'language': None,
            'registered_at': None,
            'call_status_id': None,
            'last_invoice_at': None,
            'school_sub_district': None,
            'village_district': None,
            'twitter_name': None,
            'membership_started_at': None,
            'subnations': [],
            'meetup_address': None,
            'author_id': None,
            'registered_address': None,
            'external_id': None,
            'twitter_login': None,
            'inferred_party': None,
            'spent_capital_amount_in_cents': 0,
            'suffix': None,
            'mailing_address': None,
            'is_leaderboardable': True,
            'twitter_website': None,
            'nbec_guid': None,
            'city_district': None,
            'church': None,
            'is_profile_searchable': True,
            'employer': None,
            'is_fundraiser': False,
            'email_opt_in': True,
            'recruits_count': 0,
            'email2_is_bad': False,
            'county_district': None,
            'recruiter': None,
            'twitter_friends_count': None,
            'facebook_username': None,
            'active_customer_started_at': None,
            'pf_strat_id': None,
            'locale': None,
            'twitter_address': None,
            'is_supporter': True,
            'do_not_call': False,
            'profile_image_url_ssl': 'https://d3n8a8pro7vhmx.cloudfront.net'
                                     '/assets/icons/buddy.png',
            'invoices_amount_in_cents': None,
            'username': None,
            'donations_amount_in_cents': 0,
            'is_volunteer': False,
            'civicrm_id': None,
            'supranational_district': None,
            'precinct_name': None,
            'invoice_payments_amount_in_cents': None,
            'work_phone_number': None,
            'phone': '213.394.4623',
            'received_capital_amount_in_cents': 500,
            'primary_address': None,
            'is_possible_duplicate': False,
            'invoice_payments_referred_amount_in_cents': None,
            'donations_amount_this_cycle_in_cents': 0,
            'priority_level': None,
            'first_fundraised_at': None,
            'phone_normalized': '2133944623',
            'rnc_regid': None,
            'twitter_id': None,
            'birthdate': None,
            'mobile': None,
            'federal_district': None,
            'donations_to_raise_amount_in_cents': 0,
            'support_probability_score': None,
            'invoices_count': None,
            'nbec_precinct_code': None,
            'website': None,
            'closed_invoices_amount_in_cents': None,
            'home_address': None,
            'school_district': None,
            'support_level': None,
            'demo': None,
            'children_count': 0,
            'updated_at': '2015-02-02T19:30:28-08:00',
            'membership_level_name': None,
            'billing_address': None,
            'is_ignore_donation_limits': False,
            'signup_type': 0,
            'precinct_id': None,
            'rnc_id': None,
            'id': 2,
            'ethnicity': None,
            'is_survey_question_private': False,
            'middle_name': None,
            'author': None,
            'last_fundraised_at': None,
            'state_file_id': None,
            'note': None,
            'submitted_address': None,
            'support_level_changed_at': None,
            'party': None,
            'contact_status_id': None,
            'outstanding_invoices_amount_in_cents': None,
            'page_slug': None,
            'outstanding_invoices_count': None,
            'first_recruited_at': None,
            'county_file_id': None,
            'first_name': 'Foo',
            'facebook_profile_url': None,
            'city_sub_district': None,
            'has_facebook': False,
            'is_deceased': False,
            'labour_region': None,
            'state_lower_district': None,
            'dw_id': None,
            'created_at': '2015-02-02T19:30:28-08:00',
            'is_prospect': False,
            'priority_level_changed_at': None,
            'is_mobile_bad': False,
            'overdue_invoices_count': None,
            'ngp_id': None,
            'do_not_contact': False,
            'first_donated_at': None,
            'turnout_probability_score': None
        },
        'precinct': None
    })

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_NATIONBUILDER_SLUG': 'foobar'
        })
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_NATIONBUILDER_SLUG': 'foobar'
        })
        self.do_partial_pipeline()
